/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Jan-Michael Brummer                                  *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file connection.c
 * \brief Connection handling functions
 */

#include <ffgtk.h>
#include <connection.h>

/** Global connect list pointer */
static GList *psConnectionList = NULL;
extern gchar *pnIgnore;

/**
 * \brief Add connection structure to connect list
 * \param nId connection id
 */
void connectionAddCall( int nId, int nType, const gchar *pnLocal, const gchar *pnNumber, gchar *pnName ) {
	struct sConnection *psConnection = g_malloc0( sizeof( struct sConnection ) );

	if ( psConnection != NULL ) {
		psConnection -> nId = nId;
		psConnection -> nType = nType;
		psConnection -> pnLocal = g_strdup( pnLocal );
		psConnection -> pnNumber = g_strdup( pnNumber );
		if ( pnName != NULL ) {
			psConnection -> pnName = g_strdup( pnName );
		} else {
			psConnection -> pnName = NULL;
		}
		psConnection -> psDialWindow = NULL;

		psConnectionList = g_list_append( psConnectionList, psConnection );
	}
}

/**
 * \brief Find connection entry by id
 * \param nId connection id
 * \return connection pointer or NULL on error
 */
struct sConnection *connectionFindById( int nId ) {
	GList *psList = psConnectionList;
	struct sConnection *psConnection;

	while ( psList ) {
		psConnection = psList -> data;

		if ( psConnection != NULL && psConnection -> nId == nId ) {
			return psConnection;
		}

		psList = psList -> next;
	}

	return NULL;
}

/**
 * \brief Find connection entry by number
 * \param pnNumber connection number
 * \return connection pointer or NULL on error
 */
struct sConnection *connectionFindByNumber( const gchar *pnNumber ) {
	GList *psList = psConnectionList;
	struct sConnection *psConnection;

	while ( psList ) {
		psConnection = psList -> data;

		Debug( KERN_DEBUG, "%s <-> %s\n", psConnection ? psConnection -> pnNumber : "?", pnNumber );
		if ( psConnection != NULL && !strcmp( psConnection -> pnNumber, pnNumber ) ) {
			return psConnection;
		}

		psList = psList -> next;
	}

	return NULL;
}

/**
 * \brief Add dialing window to connection information structure
 * \param psConnection connection structure
 * \param psWidget dialing window
 */
void connectionAddDialWindow( struct sConnection *psConnection, GtkWidget *psWidget ) {
	if ( psConnection != NULL ) {
		psConnection -> psDialWindow = psWidget;
	}
}

/**
 * \brief Remove connection from connection list
 * \param psConnection connection pointer
 */
void connectionRemove( struct sConnection *psConnection ) {
	if ( psConnectionList != NULL && psConnection != NULL ) {
		psConnectionList = g_list_remove( psConnectionList, psConnection );
		g_free( psConnection -> pnLocal );
		g_free( psConnection -> pnNumber );
		if ( psConnection -> pnName ) {
			g_free( psConnection -> pnName );
		}
		g_free( psConnection );
	}
}

/**
 * \brief Ignore next monitor messages
 * \param psConnection connection structure
 * */
void connectionIgnore( struct sConnection *psConnection ) {
	if ( psConnection == NULL ) {
		Debug( KERN_DEBUG, "Warning, connection == NULL\n" );
		return;
	}

	if ( pnIgnore != NULL ) {
		g_free( pnIgnore );
		pnIgnore = NULL;
	}
	pnIgnore = g_strdup( psConnection -> pnNumber );

	connectionRemove( psConnection );
	Debug( KERN_DEBUG, "Connection removed!\n" );
}
