/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Jan-Michael Brummer                                  *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file main.c
 * \brief Main entry functions
 */

#include <ffgtk.h>
#include <gio/gio.h>
#include <libgen.h>
#include <network.h>
#include <remote.h>
#include <notification.h>
#include <osdep/osdep.h>
#ifdef HAVE_DBUS
#include <bluetooth.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#endif

/** GTK main loop */
GMainLoop *psLoop = NULL;

#ifdef HAVE_DBUS
/**
 * \brief DBus signal filter
 * \param psConnection dbus connection pointer
 * \param psMessage dbus message pointer
 * \param pUserData gtk main loop pointer
 * \return dbus result code
 */
static DBusHandlerResult dbusSignalFilter( DBusConnection *psConnection, DBusMessage *psMessage, void *pUserData ) {
	GMainLoop *psLoop = pUserData;

	if ( dbus_message_is_signal( psMessage, "org.freedesktop.Local", "Disconnected" ) ) {
		g_main_loop_quit( psLoop );
		return DBUS_HANDLER_RESULT_HANDLED;
	} else if ( dbus_message_is_signal( psMessage, "org.tabos.ffgtk.dbus.Signal", "Call" ) ) {
		DBusError sError;
		gchar *pnStr = NULL;

		dbus_error_init( &sError );
		if ( dbus_message_get_args( psMessage, &sError, DBUS_TYPE_STRING, &pnStr, DBUS_TYPE_INVALID ) ) {
			g_print( "Call received: %s\n", pnStr );
			gchar *pnNumber = fullNumber( pnStr, FALSE );
			g_print( "Call received (processed): %s\n", pnNumber );
			dialNumberDialog( NULL, pnNumber );
			g_free( pnNumber );
		} else {
			g_print( "Call received, but error getting message: %s\n", sError.message );
			dbus_error_free( &sError );
		}
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

/**
 * \brief Init DBus interface
 * \param psLoop gtk main loop pointer
 * \return error code
 */
static int InitDbus( GMainLoop *psLoop ) {
	DBusConnection *psBus;
	DBusError sError;

	dbus_error_init( &sError );
	psBus = dbus_bus_get( DBUS_BUS_SESSION, &sError );
	if ( dbus_error_is_set( &sError ) ) {
		g_warning( "Failed to connect to the D-BUS daemon: %s", sError.message );
		dbus_error_free( &sError );
		return 1;
	}

	dbus_bus_add_match( psBus, "type='signal',interface='org.tabos.ffgtk.dbus.Signal'", NULL );
	dbus_connection_add_filter( psBus, dbusSignalFilter, psLoop, NULL );

	dbus_connection_setup_with_g_main( psBus, NULL );

	return 0;
}
#endif

#ifdef HAVE_FAXOPHONE
/**
 * \brief Spooler callback
 * \param psMonitor file monitor_dir
 * \param psFile file
 * \param psOtherFile other file
 * \param nEventType event type
 * \param pUserData user data
 */
static void spoolerCallback( GFileMonitor *psMonitor, GFile *psFile, GFile *psOtherFile, GFileMonitorEvent nEventType, gpointer pUserData ) {
	gchar *pnFile = g_file_get_path( psFile );
	const gchar *pnUserName = g_get_user_name();
	static int nType = -1;

	if ( pnFile == NULL || pnUserName == NULL ) {
		return;
	}

	Debug( KERN_DEBUG, "File: %s Type: %d (%s)  Event: %d (%s)\n", pnFile, 
				nType, GFileMonitorEvent2Text(nType), 
				nEventType, GFileMonitorEvent2Text( nEventType) );

	/* Sort out invalid files */
	gchar *pnName = g_strdup_printf( "%s-", pnUserName );
	if ( has_file_extension( pnFile, ".tif" ) == TRUE || strncmp( basename( pnFile ), pnName, strlen( pnName ) ) ) {
		/* Skip it */
		return;
	}


	if ( nType == -1 && nEventType == G_FILE_MONITOR_EVENT_CREATED ) {
		nType = nEventType;
	}

	if ( nEventType == G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT ) {
		if ( nType == G_FILE_MONITOR_EVENT_CREATED ) {
			nType = nEventType;
			Debug( KERN_DEBUG, _( "Print job received on spooler\n" ) );
			processFax( pnFile );
		}
		nType = -1;
	}

	g_free( pnName );
	g_free( pnFile );
}

/**
 * \brief Initialize new printer spool queue
 */
void InitSpoolWatcher( void ) {
	GFileMonitor *psMonitor = NULL;
	GFile *psFile = NULL;
	DIR *psDir = NULL;
	GError *psError = NULL;
	gchar *pnDir = g_strdup_printf( "/var/spool/ffgtk" );

	/* Check if spooler is present, create directory if needed */
	psDir = opendir( pnDir );
	if ( psDir == NULL ) {
		g_mkdir( pnDir, 0777 );
	} else {
		closedir( psDir );
	}

	Debug( KERN_DEBUG, "Setting file monitor to '%s'\n", pnDir );

	/* Create GFile for GFileMonitor */
	psFile = g_file_new_for_path( pnDir );
	/* Create file monitor for spool directory */
	psMonitor = g_file_monitor_directory( psFile, G_FILE_MONITOR_NONE, NULL, &psError );
	if ( psMonitor == NULL ) {
		Debug( KERN_WARNING, "Could not set file monitor to spooler directory. Error: %s\n", psError -> message );
	} else {
		/* Set callback for file monitor */
		g_signal_connect( psMonitor, "changed", G_CALLBACK( spoolerCallback ), NULL );
	}

	g_free( pnDir );
}
#endif

/**
 * \brief Main entry function.
 * Set window logo, localization, plugins, profiles, trayicon, server
 * \param argc argument count
 * \param argv argument vector
 * \return error code
 */
#ifdef G_OS_WIN32
int ffgtk_main(HINSTANCE hint, int argc, char **argv);
int ffgtk_main(HINSTANCE hint, int argc, char **argv) {
#else
int main( int argc, char **argv ) {
#endif
	GdkPixbuf *psLogo = NULL;

	/* Display version information */
	Debug( KERN_INFO, "ffgtk %s - %s\n", PACKAGE_VERSION, anSvnRevision );

	/* Initialize thread support */
	if ( !g_thread_supported() ) {
#if GLIB_CHECK_VERSION(2,31,0)
		/* Threads are already initialized */
#else
		g_thread_init( NULL );
#endif
#ifdef HAVE_DBUS
		dbus_g_thread_init();
#endif
		gdk_threads_init();
	}

	/* Initialize gtk */
	gtk_init( &argc, &argv );

	g_set_application_name( _( "FFGTK Router Manager" ) );

	initDirectoryPaths();

	/* Set local bindings */
	bindtextdomain( GETTEXT_PACKAGE, getLocaleDir() );

	bind_textdomain_codeset( GETTEXT_PACKAGE, "UTF-8" );
	textdomain( GETTEXT_PACKAGE );

	/* Create main loop, we will connect the dbus signal filter to this one */
	psLoop = g_main_loop_new( NULL, FALSE );

#ifdef HAVE_DBUS
	/* Init DBUS interface */
	InitDbus( psLoop );
#endif

	/* Setup program logo and set it as default window icon */
	psLogo = getTypeIcon( TYPE_FFGTK );
	if ( psLogo != NULL ) {
		gtk_window_set_default_icon( psLogo );
	}

	/* Time to load the plugins */
	gchar *pnPluginDir = getPluginDir();
	PluginsLoad( pnPluginDir );
	g_free( pnPluginDir );

	/* Time to load lookups */
	LookupLoad();
	/* Next step we load the profiles */
	ProfilesLoad();

	Debug( KERN_DEBUG, "ActionsLoad()\n" );
	/* Load actions */
	ActionsLoad();
	/* Load DTMF actions */
	DtmfActionsLoad();
	/* Load local caller list */
	LoadLocalCallerList();
	/* Add diversity network event */
	InitDiversity();
	/* Add call-monitor port network event */
	InitMonitor();

	/* Init notification system */
	InitNotification();

#ifdef HAVE_FAXOPHONE

#if 1
	/* Start old cups printer waitqueue */
	PrintServerInit();
#else
	/* Initialize new fax spool watcher */
	InitSpoolWatcher();
#endif

#ifdef HAVE_DBUS
	InitBluetooth();
#endif

	/* Initialize faxophone */
	InitFaxPhone();
#endif

	InitRemotePort();

	/* Notify complete startup */
	Debug( KERN_DEBUG, "startup complete\n" );
	gdk_notify_startup_complete();

	InitNetworkMonitor();

	/* Create tray icon and connect menu and signals */
	menuInit();

	/* Finally give control to gtk main loop */
	Debug( KERN_DEBUG, "g_main_loop_run\n" );

	g_main_loop_run( psLoop );

	if ( menuGetIcon() ) {
		g_object_unref( G_OBJECT( menuGetIcon() ) );
	}

#ifdef HAVE_FAXOPHONE
	/* Close faxophone */
	Debug( KERN_DEBUG, "CloseFaxPhone()\n" );
	CloseFaxPhone();
	Debug( KERN_DEBUG, "CloseFaxPhone() done\n" );
#endif

	/* Logout */
	if ( getActiveProfile() ) {
	Debug( KERN_DEBUG, "routerLogout()\n" );
		routerLogout( getActiveProfile() );
	}

	/* Unload plugins */
	Debug( KERN_DEBUG, "PluginsUnload()\n" );
	PluginsUnload();
	Debug( KERN_DEBUG, "PluginsUnload() done\n" );

	/* all done, get out of here */
	Debug( KERN_DEBUG, "all done\n" );

	return 0;
}
