/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Louis Lagendijk                                      *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file menu.c
 * \brief Wrapper for the menu system (trayicon and menubar)
 */

#include <ffgtk.h>
#include <trayicon.h>
#include <limits.h>
#include <osdep/osdep.h>
#include <menucallbacks.h>

#define MAX_LASTCALLS 5

/** Loading dialog */
static GtkWidget *psLoading = NULL;

static int nOpenWindows = 0;

/**
 * \brief mark windos as open
 * \param window to be marked open
 */
void markWindowOpen( int nType ) {
	nOpenWindows = nOpenWindows | nType;
	Debug( KERN_DEBUG, "nOpenWindows is now: %d\n", nOpenWindows );
}

/**
 * \brief Mark window as closed
 * \param window to be marked closed
 */
void markWindowClosed( int nType) {
	nOpenWindows = nOpenWindows & ( INT_MAX ^ nType  );
	Debug( KERN_DEBUG, "nOpenWindows is now: %d\n", nOpenWindows );
}

/**
 * \brief  Check if windows is open
 * \param window to checked
 * \return status of window
 */
bool windowIsOpen( int nType ) {
	return ( ( nOpenWindows & nType ) != 0 );
}

/**
 * \brief Get map of open windows
 * \return map of open windows
 */
int getOpenWindows( void ) {
	return nOpenWindows;
}

/**
 * \brief Initialize menu system, create menu bar and tray icon andi start tooltip thread
 */
void menuInit( void ) {
	trayIconInit();
	callMonitorInit();
}

static gboolean displayLoadingFunc( void ) {
	psLoading = gtk_message_dialog_new( NULL, GTK_DIALOG_MODAL,
		GTK_MESSAGE_INFO, GTK_BUTTONS_NONE, _( "Please wait while ffgtk loads..." ) );
	gtk_window_set_title( GTK_WINDOW( psLoading ), _( "FFGTK Loading" ) );
	gtk_window_set_position( GTK_WINDOW( psLoading ), GTK_WIN_POS_CENTER_ALWAYS );
	gtk_dialog_run( GTK_DIALOG( psLoading ) );
	return FALSE;
}

void displayLoadingDialog( void ) {
#if GTK_MAJOR_VERSION >= 3
	Debug( KERN_DEBUG, "Start loading dialog...\n" );
	g_idle_add_full(  G_PRIORITY_HIGH, ( GSourceFunc ) displayLoadingFunc, NULL, NULL ); 
#endif
}

/**
 * \brief recreate app menu
 */
void menuRecreateAppMenu( void ) {
	/* We are now ready, cancel loading dialog */
	if (psLoading != NULL) {
		Debug( KERN_DEBUG, "Removing loading dialog\n" );
		gtk_widget_hide( psLoading );
		gtk_widget_destroy( psLoading );
		psLoading = NULL;
	}
	callMonitorRecreateAppMenu();
	trayIconRecreateAppMenu();
}

/**
 * \brief set program icon on trayicon
 * \param icvon type to be set
 */
void menuSetIconType( int nType ) {
	trayIconSetIconType(nType);
}

/**
 *  \brief get menu/trayicon widget
 * \return trayicon widget
 */
GtkStatusIcon *menuGetIcon( void ) {
	return getTrayIcon();
}

/**
 * \brief Create separator item with 'text'
 * \param pnText separator text
 * \return new menuitem widget
 */
GtkWidget *createSeparatorItem( gchar *pnText ) {
        GtkWidget *psLabel = NULL;
        GtkWidget *psBox = gtk_hbox_new( FALSE, 0 );
        GtkWidget *psItem = NULL;

        psItem = gtk_menu_item_new();

        psLabel = gtk_label_new( pnText );

        gtk_box_pack_start( GTK_BOX( psBox ), gtk_hseparator_new(), TRUE, TRUE, 0 );
        gtk_box_pack_start( GTK_BOX( psBox ), psLabel, FALSE, FALSE, 0 );
        gtk_box_pack_start( GTK_BOX( psBox ), gtk_hseparator_new(), TRUE, TRUE, 0 );

        g_object_set( G_OBJECT( psItem ), "child", psBox, "sensitive", FALSE, NULL );

        return psItem;
}


/**
 * \brief Create call menu item for type nType
 * \param pnNumber call number
 * \param nType call type
 * \return new menuitem widget
 */
GtkWidget *createCallItem( char *pnNumber, int nType ) {
	GtkWidget *psLabel = NULL;
	GtkWidget *psBox = gtk_vbox_new( FALSE, 0 );
	GtkWidget *psItem = NULL;
	GtkWidget *psImage = NULL;
	GdkPixbuf *psPix = NULL;
	struct sPerson *psPerson = findPersonByNumber( pnNumber );
	gchar *pnNumberText;
#ifndef __MACOSX__
	GtkWidget *psNameLabel = NULL;
	gchar *pnName;
#endif

	psItem = gtk_image_menu_item_new();

	psLabel = gtk_label_new( "" );
#ifdef __MACOSX__
	pnNumberText = g_markup_printf_escaped( "<small>%s (%s)</small>", 
			psPerson != NULL ? psPerson -> pnDisplayName : _( "unknown" ), pnNumber );
	gtk_label_set_markup( GTK_LABEL( psLabel ), pnNumberText );
	g_free( pnNumberText );

	gtk_box_pack_start( GTK_BOX( psBox ), psLabel, FALSE, FALSE, 0 );
	gtk_misc_set_alignment( GTK_MISC( psLabel ), 0.0, 0.5 );

#else	
	pnNumberText = g_markup_printf_escaped( "<small>%s</small>", pnNumber );
	gtk_label_set_markup( GTK_LABEL( psLabel ), pnNumberText );
	g_free( pnNumberText );

	psNameLabel = gtk_label_new( "" );
	pnName = g_markup_printf_escaped( "<b><small>%s</small></b>", psPerson != NULL ? psPerson -> pnDisplayName : _( "unknown" ) );
	gtk_label_set_markup( GTK_LABEL( psNameLabel ), pnName );
	g_free( pnName );

	gtk_box_pack_start( GTK_BOX( psBox ), psNameLabel, TRUE, TRUE, 0 );
	gtk_box_pack_start( GTK_BOX( psBox ), psLabel, FALSE, FALSE, 0 );
	gtk_misc_set_alignment( GTK_MISC( psNameLabel ), 0.0, 0.5 );
	gtk_misc_set_alignment( GTK_MISC( psLabel ), 0.0, 0.5 );
#endif
	g_object_set( G_OBJECT( psItem ), "child", psBox, NULL );


	psPix = getTypeIcon( nType );
	psImage = GTK_WIDGET( gtk_image_new_from_pixbuf( psPix ) );
	gtk_image_menu_item_set_image( GTK_IMAGE_MENU_ITEM( psItem ), psImage );
	gtk_image_menu_item_set_always_show_image( GTK_IMAGE_MENU_ITEM( psItem ), TRUE );

	return psItem;
}
/**
 * \brief Add last call to submenu - callback
 * \param psMenuParentWidget connect each entry to this menu
 */
void addLastcallsMenu( GtkWidget *psMenuParentWidget ) {
	GList *psList = NULL;
	GtkWidget *psItem = NULL;
	int nCount = 0;
	GtkWidget *psSubMenu = gtk_menu_new();
	gtk_menu_item_set_submenu( GTK_MENU_ITEM( psMenuParentWidget ), psSubMenu );


	psList = psCallerList;
	if ( g_list_length( psList ) <= 0 ) {
		gtk_widget_set_sensitive( GTK_WIDGET( psMenuParentWidget ), FALSE );
		return;
	}

	psItem = createSeparatorItem( _( "Incoming" ) );
	gtk_menu_shell_append( GTK_MENU_SHELL( psSubMenu ), psItem );

	psList = psCallerList;
	nCount = 0;
	while ( psList != NULL ) {
		struct sCaller *psCaller = psList -> data;

		if ( psCaller -> nType == TYPE_CALLIN ) {
			psItem = createCallItem( psCaller -> pnNumber, 1 );
			gtk_menu_shell_append( GTK_MENU_SHELL( psSubMenu ), psItem );
			g_signal_connect( G_OBJECT( psItem ), "activate", G_CALLBACK( trayIconDialNumber ), psCaller -> pnNumber );
			gtk_widget_set_sensitive( GTK_WIDGET( psItem ), routerHasDial( getActiveProfile() ) );
			nCount++;

			if ( nCount == MAX_LASTCALLS ) {
				break;
			}
		}

		psList = psList -> next;
	}

	psItem = createSeparatorItem( _( "Outgoing" ) );
	gtk_menu_shell_append( GTK_MENU_SHELL( psSubMenu ), psItem );

	psList = psCallerList;
	nCount = 0;
	while ( psList != NULL ) {
		struct sCaller *psCaller = psList -> data;

		if ( psCaller -> nType == TYPE_CALLOUT || psCaller -> nType == TYPE_CALLOUT_NEW ) {
			psItem = createCallItem( psCaller -> pnNumber, 3 );
			gtk_menu_shell_append( GTK_MENU_SHELL( psSubMenu ), psItem );
			g_signal_connect( G_OBJECT( psItem ), "activate", G_CALLBACK( trayIconDialNumber ), psCaller -> pnNumber );
			gtk_widget_set_sensitive( GTK_WIDGET( psItem ), routerHasDial( getActiveProfile() ) );
			nCount++;

			if ( nCount == MAX_LASTCALLS ) {
				break;
			}
		}

		psList = psList -> next;
	}

	psItem = createSeparatorItem( _( "Missed" ) );
	gtk_menu_shell_append( GTK_MENU_SHELL( psSubMenu ), psItem );

	psList = psCallerList;
	nCount = 0;
	while ( psList != NULL ) {
		struct sCaller *psCaller = psList -> data;

		if ( psCaller -> nType == TYPE_CALLINFAILED ) {
			psItem = createCallItem( psCaller -> pnNumber, 2 );

			gtk_menu_shell_append( GTK_MENU_SHELL( psSubMenu ), psItem );
			g_signal_connect( G_OBJECT( psItem ), "activate", G_CALLBACK( trayIconDialNumber ), psCaller -> pnNumber );
			gtk_widget_set_sensitive( GTK_WIDGET( psItem ), routerHasDial( getActiveProfile() ) );
			nCount++;

			if ( nCount == MAX_LASTCALLS ) {
				break;
			}
		}

		psList = psList -> next;
	}
	gtk_widget_show_all( psMenuParentWidget );
}

