/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Jan-Michael Brummer                                  *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file menucallbacks.c
 * \brief contains the callbacks used in menus
 */

#include <ffgtk.h>
#include <osdep/osdep.h>

/** main gtk loop */
extern GMainLoop *psLoop;

/**
 * \brief Destroy callback
 * \param psWidget widget
 * \param pData optional data pointer
 */
gboolean destroy( GtkWidget *psWidget, gpointer pData ) {
	g_main_loop_quit( psLoop );
	return TRUE;
}

/**
 * \brief Open CallMonitor
 */
void openCallMonitor( void ) {
	if ( ! windowIsOpen( WINDOW_MONITOR ) ) {
		CallMonitor();
		menuRecreateAppMenu();
	}
}

/**
 * \brief Close CallMonitor
 */
void closeCallMonitor( void ) {
	if ( windowIsOpen (WINDOW_MONITOR ) ) {
		callMonitorHide();
	}
}

/**
 * \brief Called on tray icon activated, opens or close CallMonitor
 */
void trayIconToggled( void ) {
	if ( windowIsOpen ( WINDOW_MONITOR ) ) {
		closeCallMonitor();
	} else {
		openCallMonitor();
	}

	menuSetIconType( TYPE_CALLOUT );
}

/**
 * \brief Open addressbook - callback
 */
void trayIconAddressBook( void ) {
	AddressBook( 0 );
}

/**
 * \brief Select active profile callback
 * \param psItem menu item pressed
 * \param pUserData name of profile
 */
void trayIconProfile( GtkMenuItem *psItem, gpointer pUserData ) {
	if ( gtk_check_menu_item_get_active( GTK_CHECK_MENU_ITEM( psItem ) ) == TRUE ) {
		const gchar *pnName = pUserData;
		struct sProfile *psProfile = findProfile( pnName );
		setActiveProfile( psProfile );
		saveProfiles();
	}
}

/**
 * \brief Dial number - callback
 * \param psWidget menu item widget
 * \param pnNumber dial number
 */
void trayIconDialNumber( GtkWidget *psWidget, gchar *pnNumber ) {
	struct sPerson *psPerson = NULL;

	if ( pnNumber != NULL ) {
		psPerson = findPersonByNumber( pnNumber );
	}

	dialNumberDialog( psPerson, pnNumber );
}

/**
 * \brief Call router reconnect - callback
 */
void trayIconReconnect( void ) {
	routerReconnect( getActiveProfile() );
}

/**
 * \brief Copy external ip to clipboard - callback
 */
void trayIconIpToClipboard( void ) {
	gchar *pnIp = routerGetIp( getActiveProfile() );

	if ( pnIp != NULL ) {
		gtk_clipboard_set_text( gtk_clipboard_get( GDK_NONE ), pnIp, -1 );
		g_free( pnIp );
	}
}

/**
 * \brief Mute sounds - callback
 * \param psItem menu item widget
 * \param pUserData UNUSED
 */
void trayIconMute( GtkCheckMenuItem *psItem, gpointer pUserData ) {
	gboolean bMute = gtk_check_menu_item_get_active ( psItem ) ;

	callMonitorSetMute( getActiveProfile(), bMute );
}

/**
 * \brief Set diversity state - callback
 * \param psItem menu item widget
 * \param psDiversity diversity structure
 */
void trayIconSetDiversity( GtkMenuItem *psItem, struct sDiversity *psDiversity ) {
	psDiversity -> bSet = !psDiversity -> bSet;

	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM( psItem ), psDiversity -> bSet );
	routerSetDiversity( getActiveProfile(), psDiversity -> nIndex, psDiversity -> bSet );
}

/**
 * \brief Open help - callback
 * \param psItem menu item widget
 * \param psDiversity diversity structure
 */
void trayIconHelp( GtkMenuItem *psItem, struct sDiversity *psDiversity ) {
	osExecute( "http://www.tabos.org/forum" );
}

/**
 * \brief Open help - callback
 * \param psItem menu item widget
 * \param psDiversity diversity structure
 */
void trayIconReadme( GtkMenuItem *psItem, struct sDiversity *psDiversity ) {
	gchar *pnTmp = g_strdup_printf( "%s/README.rtf", getPkgDataDir() );
	osExecute( pnTmp );
	g_free( pnTmp );
}
