/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Jan-Michael Brummer                                  *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file monitor.c
 * \brief Call monitor handling
 */

#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <remote.h>
#include <thread.h>
#include <bluetooth.h>

#include <ffgtk.h>
#include <network.h>
#include <connection.h>
#include <notification.h>

//#define BUFFER_LENGTH 1024
//#define IMAGE_SIZE  100

/** Ignore the following number */
gchar *pnIgnore = NULL;

/** Telnet monitor callback id */
static guint nTelnetMonitorCallbackId = 0;
/** Telnet io channel */
static GIOChannel *psChannel = NULL;
/** Telnet monitor thread */
static struct sThread sMonitorThread;

/**
 * \brief Handle call
 * \param nType call type
 * \param nConnection current connection id
 * \param pnRemote remote number
 * \param pnLocal local number
 * \param pnMedium transfer medium
 */
static void HandleCall( eCallType nType, int nConnection, char *pnNumber, char *pnLocal, char *pnMedium ) {
	gchar *pnFirstName;
	gchar *pnLastName;
	gchar *pnStreet;
	gchar *pnZipCode;
	gchar *pnCity;
	struct sConnection *psConnection = NULL;
	struct sPerson *psPerson = NULL;
	struct sProfile *psProfile = getActiveProfile();
	const gchar *pnLineAccessCode = routerGetLineAccessCode( psProfile );
	gint nDisplayTime = callMonitorGetDisplayTime( psProfile );

	Debug( KERN_DEBUG, "nType %d, nConnection: %d\n", nType, nConnection );

	/* Skip line access code */
	if ( pnNumber != NULL && pnLineAccessCode != NULL && strlen( pnLineAccessCode ) > 0 ) {
		pnNumber = pnNumber + strlen( pnLineAccessCode );
	}

	/* In case we want to fast dial, remove last # */
	if ( pnNumber != NULL && pnNumber[ strlen( pnNumber ) - 1 ] == '#' ) {
		pnNumber[ strlen( pnNumber ) - 1 ] = '\0';
	}

	if ( nType == CALL_TYPE_INCOMING || nType == CALL_TYPE_OUTGOING ) {
		if ( nType == CALL_TYPE_INCOMING && pnIgnore != NULL && !g_strcmp0( pnIgnore, pnNumber ) ) {
			Debug( KERN_DEBUG, "Ignoring monitor event\n" );

			g_free( pnIgnore );
			pnIgnore = NULL;

			goto updateList;
		}

		Debug( KERN_DEBUG, "Find person by number\n" );
		psPerson = findPersonByNumber( pnNumber );

		/* Add call to connection list */
		connectionAddCall( nConnection, nType, pnLocal, pnNumber, psPerson ? psPerson -> pnDisplayName : NULL );

		Debug( KERN_DEBUG, "Checking preferences state for '%s'\n", pnLocal );
		if ( ( callMonitorGetIncomingState( psProfile, pnLocal ) && ( nType == CALL_TYPE_INCOMING ) ) || ( callMonitorGetOutgoingState( psProfile, pnLocal ) && ( nType == CALL_TYPE_OUTGOING ) ) ) {
			PlayRingtone( nType );
#ifdef HAVE_DBUS
			//bluetoothIndicateCall();
#endif

			if ( nDisplayTime != 0 ) {
				Debug( KERN_DEBUG, "Show popup\n" );
				psConnection = connectionFindById( nConnection );

				psConnection -> pNotifyWindow = showNotification( psConnection, nType, psPerson, pnNumber, pnLocal );
				Debug( KERN_DEBUG, "pNotifyWindow: %p\n", psConnection -> pNotifyWindow );

				if ( psConnection -> pNotifyWindow != NULL && psPerson == NULL &&
					( ( nType == CALL_TYPE_INCOMING && callMonitorGetReverseLookupIn( psProfile ) ) ||
					 ( nType == CALL_TYPE_OUTGOING && callMonitorGetReverseLookupOut( psProfile ) ) ) ) {
					ReverseLookup( pnNumber, &pnFirstName, &pnLastName, &pnStreet, &pnZipCode, &pnCity );
					if ( pnLastName != NULL ) {
						/* TODO: add code to handle new fields */
						updateNotification( psConnection -> pNotifyWindow, nType, pnLastName, pnStreet, pnCity );
						if ( pnStreet != NULL ) {
							g_free( pnStreet );
						}
						if ( pnCity != NULL ) {
							g_free( pnCity );
						}
					}
				}
			}
		}

		if ( nType == CALL_TYPE_OUTGOING && dialGetNumber() != NULL && !strcmp( dialGetNumber(), pnNumber ) ) {
			Debug( KERN_DEBUG, "Add connection to dial window\n" );
			connectionAddDialWindow( connectionFindById( nConnection ), dialGetDialog() );
		}
	}

	/* execute actions */
	Debug( KERN_DEBUG, "Find connection by id\n" );
	psConnection = connectionFindById( nConnection );
	if ( psConnection != NULL ) {
		Debug( KERN_DEBUG, "Found connection, if needed execute actions\n" );
		executeActions( psConnection -> nType, nType, psConnection -> pnLocal, psConnection -> pnNumber, psConnection -> pnName );
	}

	if ( nType == CALL_TYPE_CONNECT || nType == CALL_TYPE_DISCONNECT ) {
		if ( psConnection != NULL ) {
			StopRingtone();

			closeNotification( psConnection );

			/* In case we used the dial out function, update status label of dialing dialog */
			if ( ( psConnection -> nType & ~CALL_TYPE_INCOMING ) && dialGetNumber() != NULL && !strcmp( dialGetNumber(), psConnection -> pnNumber ) ) {
				if ( nType == CALL_TYPE_DISCONNECT || nType == CALL_TYPE_CONNECT ) {
					dialWindowSetStatus( psConnection -> psDialWindow, nType );
				}
			}

			if ( nType == CALL_TYPE_DISCONNECT ) {
				Debug( KERN_DEBUG, "Type: %x\n", psConnection -> nType );

				if ( psConnection -> nType == CALL_TYPE_INCOMING ) {
					/* if we missed that call, set icon to missed */
					menuSetIconType( TYPE_CALLINFAILED );
				}

				connectionRemove( psConnection );

				updateList:
				menuRecreateAppMenu();

#ifndef DEBUG_MONITOR
				g_timeout_add_seconds( 5, updateCallList, NULL );
#endif
			} else {
				psConnection -> nType |= nType;
				Debug( KERN_DEBUG, "Type set to %x\n", psConnection -> nType );
			}
		}
	}
}

void sendCall( void *pContext, void *pData ) {
	struct sCallData *psData = pContext;

	HandleCall( psData -> nType, psData -> nConnection, psData -> pnPartA, psData -> pnPartB, psData -> pnPartC );

	if ( psData -> pnPartA ) {
		g_free( psData -> pnPartA );
		psData -> pnPartA = NULL;
	}
	if ( psData -> pnPartB ) {
		g_free( psData -> pnPartB );
		psData -> pnPartB = NULL;
	}
	if ( psData -> pnPartC ) {
		g_free( psData -> pnPartC );
		psData -> pnPartC = NULL;
	}

	g_free( psData );
}

/**
 * \brief Convert fritzbox call line to a readable output
 * \param pnText input line
 * \return error code
 */
int convert( char *pnText ) {
	int anPos[ 8 ];
	int nIndex;
	int nRet = 0;
	char *pnDate = NULL;
	char *pnType = NULL;
	char *pnConnection = NULL;
	char *pnPartA = NULL;
	char *pnPartB = NULL;
	char *pnPartC = NULL;
	char *pnPartD = NULL;
	int nConnection;
	int nLen = 0;

	if ( pnText == NULL || strlen( pnText ) <= 0 ) {
		return 0;
	}
	Debug( KERN_DEBUG, "Got '%s' (%d)\n", pnText, strlen( pnText ) );

	memset( anPos, -1, sizeof( anPos ) );

	for ( nIndex = 1; nIndex < 8; nIndex++ ) {
		anPos[ nIndex ] = findString( pnText, anPos[ nIndex - 1 ] + 1, ";" );
		if ( anPos[ nIndex ] == -1 ) {
			break;
		}
	}
	anPos[ 0 ] = 0;

	nLen = anPos[ 1 ] - anPos[ 0 ];
	if ( nLen <= 0 ) {
		nRet = -1;
		goto error1;
	}

	pnDate = getSubString( pnText, anPos[ 0 ], nLen );
	if ( pnDate == NULL ) {
		Debug( KERN_DEBUG, "Could not get date string\n" );
		nRet = -2;
		goto error1;
	}

	nLen = anPos[ 2 ] - anPos[ 1 ] - 1;
	if ( nLen <= 0 ) {
		nRet = -1;
		goto error1;
	}

	pnType = getSubString( pnText, anPos[ 1 ] + 1 , nLen );
	if ( pnType == NULL ) {
		Debug( KERN_WARNING, "pnType is NULL\n" );
		nRet = -3;
		goto error1;
	}

	nLen = anPos[ 3 ] - anPos[ 2 ] - 1;
	if ( nLen <= 0 ) {
		nRet = -1;
		goto error1;
	}

	pnConnection = getSubString( pnText, anPos[ 2 ] + 1, nLen );
	if ( pnConnection == NULL ) {
		Debug( KERN_WARNING, "pnConnection is NULL\n" );
		nRet = -4;
		goto error1;
	}

	nConnection = atoi( pnConnection );

	pnPartA = anPos[ 4 ] == -1 ? NULL : getSubString( pnText, anPos[ 3 ] + 1, anPos[ 4 ] - anPos[ 3 ] - 1 );
	pnPartB = anPos[ 5 ] == -1 ? NULL : getSubString( pnText, anPos[ 4 ] + 1, anPos[ 5 ] - anPos[ 4 ] - 1 );
	pnPartC = anPos[ 6 ] == -1 ? NULL : getSubString( pnText, anPos[ 5 ] + 1, anPos[ 6 ] - anPos[ 5 ] - 1 );
	pnPartD = anPos[ 7 ] == -1 ? NULL : getSubString( pnText, anPos[ 6 ] + 1, anPos[ 7 ] - anPos[ 6 ] - 1 );

	if ( !strcmp( pnType, "CALL" ) ) {
		/**
		 * pnPartA : box port
		 * pnPartB : caller id
		 * pnPartC : called id
		 * pnPartD : POTS, SIP, ISDN, ...
		 */
		struct sCallData *psData = malloc( sizeof( struct sCallData ) );
		psData -> nType = CALL_TYPE_OUTGOING;
		psData -> nConnection = nConnection;
		psData -> pnPartA = g_strdup( pnPartC );
		psData -> pnPartB = g_strdup( pnPartB );
		psData -> pnPartC = g_strdup( pnPartD );

		remoteCallInvoke( remoteGetMainPort(), sendCall, psData, NULL );
	} else if ( !strcmp( pnType, "RING" ) ) {
		/**
		 * pnPartA : caller id
		 * pnPartB : called id
		 * pnPartC : medium
		 */
		struct sCallData *psData = malloc( sizeof( struct sCallData ) );
		psData -> nType = CALL_TYPE_INCOMING;
		psData -> nConnection = nConnection;
		psData -> pnPartA = pnPartA != NULL ? g_strdup( pnPartA ) : g_strdup( _("unknown") );
		psData -> pnPartB = g_strdup( pnPartB );
		psData -> pnPartC = g_strdup( pnPartC );

		remoteCallInvoke( remoteGetMainPort(), sendCall, psData, NULL );
	} else if ( !strcmp( pnType, "CONNECT" ) ) {
		/**
		 * pnPartA : caller id
		 * pnPartB : id
		 */
		struct sCallData *psData = malloc( sizeof( struct sCallData ) );
		psData -> nType = CALL_TYPE_CONNECT;
		psData -> nConnection = nConnection;
		psData -> pnPartA = g_strdup( pnPartA );
		psData -> pnPartB = g_strdup( pnPartB );
		psData -> pnPartC = g_strdup( pnPartC );

		remoteCallInvoke( remoteGetMainPort(), sendCall, psData, NULL );
	} else if ( !strcmp( pnType, "DISCONNECT" ) ) {
		/**
		 * pnPartA : caller id
		 * pnPartB : id
		 */
		struct sCallData *psData = malloc( sizeof( struct sCallData ) );
		psData -> nType = CALL_TYPE_DISCONNECT;
		psData -> nConnection = nConnection;
		psData -> pnPartA = g_strdup( pnPartA );
		psData -> pnPartB = g_strdup( pnPartB );
		psData -> pnPartC = g_strdup( pnPartC );

		remoteCallInvoke( remoteGetMainPort(), sendCall, psData, NULL );
	}

error1:
	if ( pnPartD ) {
		g_free( pnPartD );
	}
	if ( pnPartC ) {
		g_free( pnPartC );
	}
	if ( pnPartB ) {
		g_free( pnPartB );
	}
	if ( pnPartA ) {
		g_free( pnPartA );
	}

	if ( pnConnection ) {
		g_free( pnConnection );
	}

	if ( pnType ) {
		g_free( pnType );
	}

	if ( pnDate ) {
		g_free( pnDate );
	}

	return nRet;
}

gboolean monitorCallback( GIOChannel *psSource, GIOCondition nCondition, gpointer pData ) {
	GIOStatus nRet;
	GError *psError = NULL;
	gchar *pnMsg;
	gsize nLen;

	switch ( nCondition ) {
		case G_IO_IN:
		case G_IO_PRI:
			Debug( KERN_DEBUG, "Juhuu, data received\n" );
			nRet = g_io_channel_read_line( psSource, &pnMsg, &nLen, NULL, &psError );
			if ( nRet == G_IO_STATUS_ERROR ) {
				g_error( "Error reading: %s\n", psError -> message );
			}

			gchar **ppnLines = g_strsplit( pnMsg, "\n", -1 );
			if ( ppnLines != NULL ) {
				gint nCount = 0;
				while ( ppnLines[ nCount ] != NULL ) {
					Debug( KERN_DEBUG, "Convert line: %s\n", ppnLines[ nCount ] );
					convert( ppnLines[ nCount ] );
					nCount++;
				}
				g_strfreev( ppnLines );
			}
			g_free( pnMsg );
			break;
		case G_IO_ERR:
		case G_IO_HUP:
			Debug( KERN_DEBUG, "Errrrr, damn it\n" );
			g_error( "Read end of pipe died!\n" );
			break;
		default:
			Debug( KERN_DEBUG, "Unhandled case: %d\n", nCondition );
			break;
	}
	return TRUE;
}

/**
 * \brief Telnet monitor socket callback
 * \param psSource source channel pointer
 * \param nCondition io condition
 * \param pData user data (UNUSED)
 * \return TRUE
 */
gboolean telnetMonitorCallback( GIOChannel *psSource, GIOCondition nCondition, gpointer pData ) {
	GIOStatus nRet;
	GError *psError = NULL;
	gchar *pnMsg;
	gsize nLen;

	switch ( nCondition ) {
		case G_IO_IN:
		case G_IO_PRI:
			nRet = g_io_channel_read_line( psSource, &pnMsg, &nLen, NULL, &psError );
			if ( nRet != G_IO_STATUS_NORMAL ) {
				g_error( "Error reading: %s\n", psError -> message );
			}

			gchar **ppnLines = g_strsplit( pnMsg, "\n", -1 );
			if ( ppnLines != NULL ) {
				gint nCount = 0;
				while ( ppnLines[ nCount ] != NULL ) {
					Debug( KERN_DEBUG, "Convert line: %s\n", ppnLines[ nCount ] );
					convert( ppnLines[ nCount ] );
					nCount++;
				}
				g_strfreev( ppnLines );
			}
			g_free( pnMsg );
			break;
		case G_IO_ERR:
		case G_IO_HUP:
			g_error( "Read end of pipe died!\n" );
			break;
		default:
			Debug( KERN_DEBUG, "Unhandled case: %d\n", nCondition );
			break;
	}

	return TRUE;
}

/**
 * \brief Initialize server and wait for incoming transmissions
 * \param data data pointer
 * \return error code
 */
gpointer telnetMonitor( gpointer data ) {
	GSocket *psSocket = NULL;
	GInetAddress *psInetAddress = NULL;
	GSocketAddress *psSockAddress = NULL;
	GError *psError = NULL;
	GResolver *psResolver = NULL;
	GList *psList = NULL;
	int nSock;
	const gchar *pnHostName = NULL;
	GList *psIter;

	do {
		nSock = -1;

		if ( getActiveProfile() == NULL ) {
			Debug( KERN_DEBUG, "No active profile\n" );
			goto end;
		}

		pnHostName = routerGetHost( getActiveProfile() );
		if ( pnHostName == NULL || strlen( pnHostName ) <= 0 ) {
			Debug( KERN_DEBUG, "Invalid hostname\n" );
			goto end;
		}

		Debug( KERN_DEBUG, "Trying to connect to '%s' on port 1012\n", pnHostName );

		psResolver = g_resolver_get_default();
		psList = g_resolver_lookup_by_name( psResolver, pnHostName, NULL, NULL );
		g_object_unref( psResolver );

		if ( psList == NULL ) {
			Debug( KERN_DEBUG, "Cannot resolve ip from hostname!\n" );
			goto end;
		}

		for ( psIter = psList; psIter != NULL; psIter = psIter -> next ) {
			psInetAddress = psIter -> data;

			if ( g_inet_address_get_family( psInetAddress ) == G_SOCKET_FAMILY_IPV4 ) {
				break;
			} else {
				psInetAddress = NULL;
			}
		}

		if ( psInetAddress == NULL ) {
			Debug( KERN_WARNING, "Could not get required IPV4 connection to telnet port!\n" );
			goto end;
		}


		Debug( KERN_DEBUG, "IP: %s\n", g_inet_address_to_string( psInetAddress ) );


		if ( psInetAddress == NULL ) {
			Debug( KERN_DEBUG, "Could not get inet address from string: '%s'\n", pnHostName );
			g_object_unref( psSocket );
			g_resolver_free_addresses( psList );
			goto end;
		}

		psSockAddress = g_inet_socket_address_new( psInetAddress, 1012 );
		if ( psSockAddress == NULL ) {
			Debug( KERN_DEBUG, "Could not create sock address on port 1012\n" );
			g_object_unref( psSocket );
			g_resolver_free_addresses( psList );
			goto end;
		}

		psError = NULL;
		psSocket = g_socket_new( g_inet_address_get_family( psInetAddress ), G_SOCKET_TYPE_STREAM, G_SOCKET_PROTOCOL_TCP, &psError );
		if ( psError != NULL ) {
			Debug( KERN_DEBUG, "Could not create socket. Error: '%s'\n", psError -> message );
			g_error_free( psError );
			goto end;
		}

		psError = NULL;
		if ( g_socket_connect( psSocket, psSockAddress, NULL, &psError ) == FALSE ) {
			Debug( KERN_DEBUG, "Could not connect to socket. Error: %s\n", psError -> message );
			g_error_free( psError );
			g_object_unref( psSocket );
			g_resolver_free_addresses( psList );
			goto end;
		}

		Debug( KERN_DEBUG, "Connected to '%s' on port 1012\n", g_inet_address_to_string( psInetAddress ) );

		nSock = g_socket_get_fd( psSocket );

	#ifdef G_OS_WIN32
		psChannel = g_io_channel_win32_new_socket( nSock );
	#else
		psChannel = g_io_channel_unix_new( nSock );
	#endif
		g_io_channel_set_encoding( psChannel, NULL, NULL );
		nTelnetMonitorCallbackId = g_io_add_watch( psChannel, G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP, telnetMonitorCallback, NULL );

		g_resolver_free_addresses( psList );

		return NULL;
	end:
		g_usleep( 2 * G_USEC_PER_SEC );
	} while ( !ThreadIsStopping( &sMonitorThread ) );

	return NULL;
}

/**
 * \brief Telnet monitor connect
 * \return error code
 */
int telnetMonitorConnect( void ) {
	Debug( KERN_DEBUG, "Called\n" );

	if ( !ThreadIsRunning( &sMonitorThread ) ) {
		InitThread( &sMonitorThread, "Monitor Connect Thread" );
		StartThread( &sMonitorThread, telnetMonitor, NULL );
		return 0;
	}

	return -1;
}

/**
 * \brief Telnet monitor disconnect
 * \return error code
 */
int telnetMonitorDisconnect( void ) {
	Debug( KERN_DEBUG, "Called\n" );

	if ( ThreadIsRunning( &sMonitorThread ) ) {
		StopThread( &sMonitorThread );
	}

	if ( nTelnetMonitorCallbackId != 0 ) {
		Debug( KERN_DEBUG, "Removing monitor callback\n" );
		g_source_remove( nTelnetMonitorCallbackId );
		nTelnetMonitorCallbackId = 0;
	}

	if ( psChannel != NULL ) {
		Debug( KERN_DEBUG, "IO channel shutdown\n" );
		g_io_channel_shutdown( psChannel, FALSE, NULL );
		Debug( KERN_DEBUG, "IO channel unref\n" );
		g_io_channel_unref( psChannel );
		psChannel = NULL;
	}

	return 0;
}

/**
 * \brief Starts call monitor thread
 */
void InitMonitor( void ) {
	AddNetworkEvent( telnetMonitorConnect, telnetMonitorDisconnect );
}
