/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Jan-Michael Brummer                                  *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file network.c
 * \brief Network state monitor
 */

#include <ffgtk.h>
#include <network.h>
#include <thread.h>

/** Online state flag */
static gboolean bOnline = FALSE;
/** Network event list */
static GList *psNetworkEventList = NULL;

/**
 * \brief Add new network event handlers
 * \param pConnect network connect function
 * \param pDisconnect network disconnect function
 */
void AddNetworkEvent( networkConnect pConnect, networkDisconnect pDisconnect ) {
	struct sNetworkEvent *psEvent = NULL;

	g_assert( pConnect != NULL );
	g_assert( pDisconnect != NULL );

	/* Allocate new structure and set pointers */
	psEvent = g_malloc( sizeof( struct sNetworkEvent ) );

	psEvent -> Connect = pConnect;
	psEvent -> Disconnect = pDisconnect;

	/* Add to network event list */
	psNetworkEventList = g_list_append( psNetworkEventList, psEvent );

	if ( bOnline == TRUE ) {
		psEvent -> Connect();
	}
}

/**
 * \brief State change callback - loop through all registered events and call handlers
 * \param nState 0 = offline, 1 = online
 * \return error code
 */
static int networkStateChanged( guint nState ) {
	GList *psList = NULL;
	int nRet = 0;

	Debug( KERN_DEBUG, "called\n" );

	/* Loop through network event list */
	for ( psList = psNetworkEventList; psList != NULL && psList -> data != NULL; psList = psList -> next ) {
		struct sNetworkEvent *psEvent = psList -> data;

		switch ( nState ) {
		case 0:
			/* If available, call disconnect */
			if ( psEvent -> Disconnect != NULL ) {
				psEvent -> Disconnect();
			}
			break;
		case 1:
			/* If available, call connect */
			if ( psEvent -> Connect != NULL ) {
				psEvent -> Connect();
			}
			break;
		default:
			Debug( KERN_DEBUG, "Unhandled state change (%d)\n", nState );
			nRet = -1;
			break;
		}

		g_usleep( G_USEC_PER_SEC * 0.5f );
	}

	return nRet;
}


//#if GLIB_CHECK_VERSION(2, 32, 0)
/**
 * \brief Network monitor signals state change
 * \param psMonitor network monitor
 * \param bAvailable the current value of network-available
 * \param pUserData user data (UNUSED)
 */
//static void networkMonitorChanged( GNetworkMonitor *psMonitor, gboolean bAvailable, gpointer pUserData ) {
//	networkStateChanged( bAvailable );
//}

//#else

/**
 * \brief Ping router function
 * \param pData data pointer (UNUSED)
 * \return TRUE (always stay alive)
 */
static gpointer pingRouter( gpointer pData ) {
	struct sProfile *psProfile = getActiveProfile();
	struct sUrlHandler *psHandler = NULL;
	int nError;

	while ( 1 ) {
	if ( psProfile == NULL ) {
		g_usleep( 5 * G_USEC_PER_SEC );
		continue;
	}

	psHandler = urlHandler( routerGetHost( psProfile ), 80 );

	/* Currently try to load standard page - TODO: switch to header only! */
	nError = readUrl( psHandler, NULL );

	/* Check error code and detect network state change */
	if ( nError == CURLE_OK && bOnline == FALSE ) {
		/* Connected to router */
		networkStateChanged( 1 );
	} else if ( nError != CURLE_OK && bOnline == TRUE ) {
		/* Diconnected */
		networkStateChanged( 0 );
	}

	/* Update online state */
	bOnline = ( nError == CURLE_OK );

	freeHandler( psHandler );
		g_usleep( 5 * G_USEC_PER_SEC );
	}

	return NULL;
}
//#endif

/**
 * \brief Initialize router check function as idle loop
 * \return error code
 */
int InitNetworkMonitor( void ) {
//#if GLIB_CHECK_VERSION(2, 32, 0)
//	GNetworkMonitor *psMonitor = g_network_monitor_get_default();
//
//	g_signal_connect( psMonitor, "network-changed", G_CALLBACK( networkMonitorChanged ), NULL );
//
//	return 0;
//#else
	/* Faster startup */
	//pingRouter( NULL );

	//return g_timeout_add_seconds( 5, pingRouter, NULL );

	struct sThread sNetworkThread;

	InitThread( &sNetworkThread, "Network test" );
	StartThread( &sNetworkThread, pingRouter, NULL );

	return 0;
//#endif
}
