/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Jan-Michael Brummer                                  *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file notification-gtk.c
 * \brief Notification window based on GTK
 */

#include <ffgtk.h>
#include <connection.h>
#include <notification.h>

/**
 * \brief Show popup call notification (GTK)
 * \param nType call type
 * \param psPerson person structure
 * \param pnNumber remote number
 * \param pnLocal local number
 * \return popup window
 */
static void *showNotificationGtk( struct sConnection *psConnection, eCallType nType, struct sPerson *psPerson, gchar *pnNumber, gchar *pnLocal ) {
	GtkBuilder *psBuilder = NULL;
	GError *psError = NULL;
	GtkWidget *psDialog;
	GtkWidget *psFrame;
	gint nWidth, nHeight;
	gint nPosition = callMonitorGetPopupPosition( getActiveProfile() );
	GtkWidget *psTypeLabel;
	GtkWidget *psLocalLabel;
	GtkWidget *psDirectionLabel;
	GtkWidget *psNameLabel;
	GtkWidget *psCompanyLabel;
	GtkWidget *psStreetLabel;
	GtkWidget *psCityLabel;
	GtkWidget *psImage;
	GtkWidget *psAcceptButton;
	GtkWidget *psDenyButton;
	gchar *pnUiFile = getUiFile( "popup.ui" );
	GdkPixbuf *psScaled;
	gchar *pnTmp = NULL;
	gint nMenuBarOffset = 0;

	psBuilder = gtk_builder_new();
	if ( gtk_builder_add_from_file( psBuilder, pnUiFile, &psError ) == 0 ) {
		Debug( KERN_WARNING, "Error: %s\n", psError -> message );
		g_error_free( psError );
		g_free( pnUiFile );
		return NULL;
	}
	g_free( pnUiFile );

	Debug( KERN_DEBUG, "Showing popup\n" );
	psDialog = GTK_WIDGET( gtk_builder_get_object( psBuilder, "psPopup" ) );
	psFrame = GTK_WIDGET( gtk_builder_get_object( psBuilder, "psFrame" ) );
	psTypeLabel = GTK_WIDGET( gtk_builder_get_object( psBuilder, "psTypeLabel" ) );
	psLocalLabel = GTK_WIDGET( gtk_builder_get_object( psBuilder, "psLocalLabel" ) );
	psDirectionLabel = GTK_WIDGET( gtk_builder_get_object( psBuilder, "psDirectionLabel" ) );
	psNameLabel = GTK_WIDGET( gtk_builder_get_object( psBuilder, "psNameLabel" ) );
	psCompanyLabel = GTK_WIDGET( gtk_builder_get_object( psBuilder, "psCompanyLabel" ) );
	psStreetLabel = GTK_WIDGET( gtk_builder_get_object( psBuilder, "psStreetLabel" ) );
	psCityLabel = GTK_WIDGET( gtk_builder_get_object( psBuilder, "psCityLabel" ) );
	psImage = GTK_WIDGET( gtk_builder_get_object( psBuilder, "psImage" ) );
	psAcceptButton = GTK_WIDGET( gtk_builder_get_object( psBuilder, "psAcceptButton" ) );
	psDenyButton = GTK_WIDGET( gtk_builder_get_object( psBuilder, "psDenyButton" ) );

#if defined HAVE_FAXOPHONE
	g_signal_connect( G_OBJECT( psAcceptButton ), "clicked", G_CALLBACK( acceptClicked ), psConnection );
	g_signal_connect( G_OBJECT( psDenyButton ), "clicked", G_CALLBACK( denyClicked ), psConnection );
#else
	gtk_widget_set_sensitive( GTK_WIDGET( psAcceptButton ), FALSE );
	gtk_widget_set_sensitive( GTK_WIDGET( psDenyButton ), FALSE );
#endif

	g_object_set_data( G_OBJECT( psDialog ), "type_label", psTypeLabel );
	g_object_set_data( G_OBJECT( psDialog ), "direction_label", psDirectionLabel );
	g_object_set_data( G_OBJECT( psDialog ), "name_label", psNameLabel );
	g_object_set_data( G_OBJECT( psDialog ), "street_label", psStreetLabel );
	g_object_set_data( G_OBJECT( psDialog ), "city_label", psCityLabel );
	g_object_set_data( G_OBJECT( psDialog ), "number" , pnNumber ? strdup( pnNumber ) : strdup( "" ) );
	g_object_set_data( G_OBJECT( psDialog ), "person" , psPerson );

	if ( nType == CALL_TYPE_INCOMING ) {
		gtk_label_set_markup( GTK_LABEL( psTypeLabel ), _( "<b>Incoming call</b>" ) );
		if ( pnLocal != NULL && strlen( pnLocal ) > 0 ) {
			pnTmp = g_strdup_printf( _( "(on %s)" ), pnLocal );
		} else {
			pnTmp = g_strdup_printf( _( "(on ?)" ) );
		}
		gtk_label_set_text( GTK_LABEL( psLocalLabel ), pnTmp );
		g_free( pnTmp );
		gtk_label_set_markup( GTK_LABEL( psDirectionLabel ), _( "<i>From:</i>" ) );

		gtk_widget_show( GTK_WIDGET( psAcceptButton ) );
		gtk_widget_show( GTK_WIDGET( psDenyButton ) );
	} else {
		gtk_label_set_markup( GTK_LABEL( psTypeLabel ), _( "<b>Outgoing call</b>" ) );
		gtk_label_set_markup( GTK_LABEL( psDirectionLabel ), _( "<i>To:</i>" ) );
	}

	gtk_label_set_text( GTK_LABEL( psStreetLabel ), _( "Unknown" ) );
	gtk_label_set_text( GTK_LABEL( psCityLabel ), _( "Unknown" ) );

	if ( psPerson == NULL ) {
		gtk_label_set_text( GTK_LABEL( psNameLabel ), pnNumber );
	} else {
		gtk_label_set_text( GTK_LABEL( psNameLabel ), psPerson -> pnDisplayName );
		gtk_label_set_text( GTK_LABEL( psCompanyLabel ), psPerson -> pnCompany );
		gtk_label_set_text( GTK_LABEL( psStreetLabel ), psPerson -> pnPrivateStreet );
		gchar *pnTmp = g_strdup_printf( "%s %s", psPerson -> pnPrivateZipCode, psPerson -> pnPrivateCity );
		gtk_label_set_text( GTK_LABEL( psCityLabel ), pnTmp );
		g_free( pnTmp );

		psScaled = getScaledImage( psPerson, 2 );
		if ( psScaled != NULL ) {
			gtk_image_set_from_pixbuf( GTK_IMAGE( psImage ), psScaled );
		}
	}

	gtk_widget_show_all( GTK_WIDGET( psFrame ) );

	gtk_window_get_size( GTK_WINDOW( psDialog ), &nWidth, &nHeight );

#ifdef __MACOSX__
	nMenuBarOffset = 24;
#endif

#ifdef G_OS_WIN32
	nHeight += 32;
#endif

	switch ( nPosition ) {
		case 0:
			/* Top Left */
			gtk_window_move( GTK_WINDOW( psDialog ), 0, nMenuBarOffset );
			break;
		case 1:
			/* Top Right */
			gtk_window_move( GTK_WINDOW( psDialog ), gdk_screen_width() - nWidth, nMenuBarOffset );
			break;
		case 2:
			/* Bottom Right */
			gtk_window_move( GTK_WINDOW( psDialog ), gdk_screen_width() - nWidth, gdk_screen_height() - nHeight );
			break;
		case 3:
			/* Bottom Left */
			gtk_window_move( GTK_WINDOW( psDialog ), 0, gdk_screen_height() - nHeight );
			break;
	}

	gtk_builder_connect_signals( psBuilder, NULL );

	g_object_unref( G_OBJECT( psBuilder ) );

	/* Timeout sec * 1000 */
	psConnection -> pNotifyWindow = psDialog;
	if ( nType == CALL_TYPE_OUTGOING ) {
		psConnection -> nNotifyTimeout = g_timeout_add_seconds( callMonitorGetDisplayTime( getActiveProfile() ), closeNotification, psConnection );
	}

	gtk_widget_show_all( GTK_WIDGET( psDialog ) );

	gtk_window_stick( GTK_WINDOW( psDialog ) );
	gtk_window_set_keep_above( GTK_WINDOW( psDialog ), TRUE );

	return psDialog;
}

/**
 * \brief Update person information in popup window (GTK)
 * \param psPopup popup window
 * \param nType call nype
 * \param pnName remote name
 * \param pnStreet remote street
 * \param pnCity remote city
 */
static void updateNotificationGtk( void *psPopup, eCallType nType, gchar *pnName, gchar *pnStreet, gchar *pnCity ) {
	GtkWidget *psTypeLabel = g_object_get_data( G_OBJECT( psPopup ), "type_label" );
	GtkWidget *psDirectionLabel = g_object_get_data( G_OBJECT( psPopup ), "direction_label" );
	GtkWidget *psNameLabel = g_object_get_data( G_OBJECT( psPopup ), "name_label" );
	GtkWidget *psStreetLabel = g_object_get_data( G_OBJECT( psPopup ), "street_label" );
	GtkWidget *psCityLabel = g_object_get_data( G_OBJECT( psPopup ), "city_label" );
	gint nWidth, nHeight;
	gint nPosition = callMonitorGetPopupPosition( getActiveProfile() );

	Debug( KERN_DEBUG, "Updating popup\n" );
	gtk_widget_hide( GTK_WIDGET( psPopup ) );

	gtk_label_set_text( GTK_LABEL( psTypeLabel ), nType == CALL_TYPE_INCOMING ? _( "Incoming call" ) : _( "Outgoing call" ) );
	gtk_label_set_text( GTK_LABEL( psDirectionLabel ), nType == CALL_TYPE_INCOMING ? _( "From:" ) : _( "To:" ) );
	gtk_label_set_text( GTK_LABEL( psNameLabel ), pnName );
	if ( pnStreet != NULL ) {
		gtk_label_set_text( GTK_LABEL( psStreetLabel ), pnStreet );
	}
	if ( pnCity != NULL ) {
		gtk_label_set_text( GTK_LABEL( psCityLabel ), pnCity );
	}
	g_object_set_data( G_OBJECT( psPopup ), "name" , pnName ? strdup( pnName ) : strdup( "" ) );

	gtk_window_get_size( GTK_WINDOW( psPopup ), &nWidth, &nHeight );

	switch ( nPosition ) {
		case 0:
			gtk_window_move( GTK_WINDOW( psPopup ), 0, 0 );
			break;
		case 1:
			gtk_window_move( GTK_WINDOW( psPopup ), gdk_screen_width() - nWidth, 0 );
			break;
		case 2:
			gtk_window_move( GTK_WINDOW( psPopup ), gdk_screen_width() - nWidth, gdk_screen_height() - nHeight );
			break;
		case 3:
			gtk_window_move( GTK_WINDOW( psPopup ), 0, gdk_screen_height() - nHeight );
			break;
	}

	gtk_widget_show_all( GTK_WIDGET( psPopup ) );
}

/**
 * \brief Close notification window
 * \param pWindow pointer to notify window
 */
static void closeNotificationGtk( gpointer pWindow ) {
	Debug( KERN_DEBUG, "Close popup\n" );
	gtk_widget_destroy( GTK_WIDGET( pWindow ) );
}

/**
 * \brief Initialize notification system
 * \return notification structure or NULL on error
 */
struct sNotification *InitNotificationGtk( void ) {
	struct sNotification *psNotification;

	psNotification = malloc( sizeof( struct sNotification ) );
	if ( psNotification != NULL ) {
		Debug( KERN_DEBUG, "Using GTK notification\n" );
		psNotification -> CloseNotification = closeNotificationGtk;
		psNotification -> UpdateNotification = updateNotificationGtk;
		psNotification -> ShowNotification = showNotificationGtk;
	}

	return psNotification;
}
