/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Jan-Michael Brummer                                  *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file notification-notify.c
 * \brief Notification handling with libnotify
 */

#include <ffgtk.h>
#include <connection.h>
#include <notification.h>

#include <libnotify/notify.h>

/**
 * \brief Notify accept clicked - wrapper to acceptClicked()
 * \param psNotification notification window
 * \param pnAction clicked action
 * \param pUserData user data - phone number
 */
static void notifyAcceptClicked( NotifyNotification *psNotification, gchar *pnAction, gpointer pUserData ) {
	acceptClicked( NULL, pUserData );
}

/**
 * \brief Deny incoming call
 * \param psWidget button widget
 * \param pData popup widget pointer
 */
static void notifyDenyClicked( NotifyNotification *psNotification, gchar *pnAction, gpointer pUserData ) {
	denyClicked( NULL, pUserData );
}

/**
 * \brief Show popup call notification (libnotify)
 * \param nType call type
 * \param psPerson person structure
 * \param pnNumber remote number
 * \param pnLocal local number
 * \return popup window
 */
static void *showNotificationNotify( struct sConnection *psConnection, eCallType nType, struct sPerson *psPerson, gchar *pnNumber, gchar *pnLocal ) {
#ifndef HAVE_APPINDICATOR
	GdkScreen *psScreen = NULL;
	GdkRectangle sArea;
#endif
	gchar *pnText = NULL;
	NotifyNotification *psTooltip = NULL;

#if NOTIFY_VERSION_MINOR >= 7
	notify_init( "ffgtk" );
#endif

	Debug( KERN_DEBUG, "Show notification\n" );

	if ( nType == CALL_TYPE_INCOMING ) {
		if ( psPerson != NULL ) {
			pnText = g_markup_printf_escaped( _( "From:\t\t%s\nCompany:\t%s\nStreet:\t\t%s\nCity:\t\t%s %s\n" ),
				psPerson -> pnDisplayName,
				psPerson -> pnCompany,
				psPerson -> pnPrivateStreet,
				psPerson -> pnPrivateZipCode,
				psPerson -> pnPrivateCity );
		} else {
			pnText = g_markup_printf_escaped( _( "<i>From:</i> %s" ), pnNumber );
		}
	} else {
		if ( psPerson != NULL ) {
			pnText = g_markup_printf_escaped( _( "To:\t\t%s\nCompany:\t%s\nStreet:\t\t%s\nCity:\t\t%s %s\n" ),
				psPerson -> pnDisplayName,
				psPerson -> pnCompany,
				psPerson -> pnPrivateStreet,
				psPerson -> pnPrivateZipCode,
				psPerson -> pnPrivateCity );
		} else {
			pnText = g_markup_printf_escaped( _( "<i>To:</i> %s" ), pnNumber );
		}
	}

	if ( psPerson != NULL && psPerson -> psImage != NULL ) {
		psTooltip = notify_notification_new( nType == CALL_TYPE_INCOMING ? _( "Incoming call" ) : _( "Outgoing call" ), pnText, NULL
#if NOTIFY_VERSION_MINOR < 7
	, NULL
#endif
	);
		notify_notification_set_icon_from_pixbuf( psTooltip, psPerson -> psImage );
	} else {
		psTooltip = notify_notification_new( nType == CALL_TYPE_INCOMING ? _( "Incoming call" ) : _( "Outgoing call" ), pnText,
			nType == CALL_TYPE_INCOMING ? /*SHAREDIR*/ "notification-message-ffgtk-in.svg" : /*SHAREDIR*/ "notification-message-ffgtk-out.svg"
#if NOTIFY_VERSION_MINOR < 7
			, NULL
#endif
		);
	}

#if defined HAVE_FAXOPHONE
	if ( nType == CALL_TYPE_INCOMING ) {
		notify_notification_add_action( psTooltip, "accept", _( "Accept" ), notifyAcceptClicked, psConnection, NULL );
		notify_notification_add_action( psTooltip, "deny", _( "Decline" ), notifyDenyClicked, psConnection, NULL );
	}
#endif

	notify_notification_set_category( psTooltip, "information" );
	notify_notification_set_timeout( psTooltip, NOTIFY_EXPIRES_NEVER );
	notify_notification_set_urgency( psTooltip, NOTIFY_URGENCY_CRITICAL );
#if NOTIFY_VERSION_MINOR < 7
	notify_notification_attach_to_status_icon( psTooltip, menuGetIcon() );
#endif

#ifndef HAVE_APPINDICATOR
	gtk_status_icon_get_geometry( menuGetIcon(), &psScreen, &sArea, NULL );
#if NOTIFY_VERSION_MINOR < 7
	notify_notification_set_geometry_hints( psTooltip, psScreen, sArea.x, sArea.y );
#endif
#endif
	notify_notification_show( psTooltip, NULL );

	g_free( pnText );

	psConnection -> pNotifyWindow = psTooltip;

	if ( nType == CALL_TYPE_OUTGOING ) {
		psConnection -> nNotifyTimeout = g_timeout_add_seconds( callMonitorGetDisplayTime( getActiveProfile() ), closeNotification, psConnection );
	}

	return psTooltip;
}

/**
 * \brief Update person information in popup window (System)
 * \param psPopup popup window
 * \param nType call type
 * \param pnName remote name
 * \param pnStreet remote street
 * \param pnCity remote city
 */
static void updateNotificationNotify( void *psTooltip, eCallType nType, gchar *pnName, gchar *pnStreet, gchar *pnCity ) {
	gchar *pnText = NULL;

	Debug( KERN_DEBUG, "Update notification\n" );

	pnText = g_markup_printf_escaped( nType == CALL_TYPE_INCOMING ? _( "<i>From:</i> %s" ) : _( "<i>To:</i> %s" ), pnName );
	notify_notification_update( psTooltip, nType == CALL_TYPE_INCOMING ? _( "Incoming call" ) : _( "Outgoing call" ), pnText, "gtk-dialog-info" );

	notify_notification_show( psTooltip, NULL );
	g_free( pnText );
}

/**
 * \brief Close notification window
 * \param pWindow notifyication window
 */
static void closeNotificationNotify( gpointer pWindow ) {
	Debug( KERN_DEBUG, "Closing notification\n" );
	notify_notification_close( pWindow, NULL );
}

/**
 * \brief Test if notification server has persistance
 * \return boolean
 */

gboolean notificationNotifyHasPersistance ( void ) {
	gboolean bHasPersistance = FALSE;
	GList *psCaps = notify_get_server_caps ();

	if (psCaps == NULL) {
		Debug( KERN_DEBUG, "Failed to receive server capabilities\n" );
		return FALSE;
	}

	bHasPersistance = ( g_list_find_custom ( psCaps, "persistence", (GCompareFunc)strcmp) != NULL);
	g_list_foreach ( psCaps, (GFunc) g_free, NULL);
	g_list_free ( psCaps );
	
	return bHasPersistance;
}

/**
 * \brief Initialize libnotify notifications
 * \return notification pointer or NULL on error
 */
struct sNotification *InitNotificationNotify( void ) {
	struct sNotification *psNotification;
	GList *psCapabilities = notify_get_server_caps();
	GList *psCap;
	gboolean bAcceptActions = FALSE;

	if ( psCapabilities != NULL ) {
		for ( psCap = psCapabilities; psCap != NULL; psCap = psCap -> next ) {
			if ( strcmp( ( char * ) psCap -> data, "actions" ) == 0 ) {
				bAcceptActions = TRUE;
				break;
			}
		}

		g_list_foreach( psCapabilities, (GFunc)g_free, NULL );
		g_list_free( psCapabilities );
	}

	if ( bAcceptActions == FALSE ) {
		Debug( KERN_DEBUG, "libnotify does not support actions, so we will not use this notification system\n" );
		return NULL;
	}
	Debug( KERN_INFO, "libnotify accepts actions, so use libnotify!\n" );

	psNotification = malloc( sizeof( struct sNotification ) );
	if ( psNotification != NULL ) {
		psNotification -> CloseNotification = closeNotificationNotify;
		psNotification -> UpdateNotification = updateNotificationNotify;
		psNotification -> ShowNotification = showNotificationNotify;
	}

	return psNotification;
}
