/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Jan-Michael Brummer                                  *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file remote.c
 * \brief Remote call functions
 */

#include <ffgtk.h>
#include <remote.h>

/** Main remote call port */
static struct sRemoteCallPort *psMainPort;

/**
 * \brief Process remote call request
 * \param psPort remote call port
 * \param psEvent remote call event
 * \return 0
 */
static int remoteCallProcess( struct sRemoteCallPort *psPort, struct sRemoteCall *psEvent ) {
	psEvent -> pFunc( psEvent -> pContext, psEvent -> pData );

	if ( psPort -> psCondition ) {
		/* Wake up the thread which singalled main */
		g_mutex_lock( psPort -> psMutex );
		g_cond_signal( psPort -> psCondition );
		g_mutex_unlock( psPort -> psMutex );
	}

	return 0;
}

/**
 * \brief Handle remote call port
 * \param psChannel io channel
 * \param nCondition io condition
 * \param pData user data pointer to remote call port structure
 * \\eturn TRUE
 */
gboolean handleRemoteCall( GIOChannel *psChannel, GIOCondition nCondition, gpointer pData ) {
	struct sRemoteCallPort *psPort = pData;
	struct sRemoteCall sEvent;
	GIOStatus nStatus;
	GError *psError = NULL;
	gsize nRead;

	/* Read one remote call item from pipe */
	nStatus = g_io_channel_read_chars( psPort -> psChannelIn, ( gchar * ) &sEvent, sizeof( struct sRemoteCall ), &nRead, &psError );
	if ( nStatus == G_IO_STATUS_NORMAL && nRead == sizeof( sEvent ) ) {
		remoteCallProcess( psPort, &sEvent );
	} else {
		Debug( KERN_WARNING, "Warning, nStatus = %d, nRead = %d\n", nStatus, nRead );
	}

	return TRUE;
}

/**
 * \brief Initialize remote call port
 * \param psPort remote call port
 * \return error code (0 on success, otherwise error)
 */
int remoteCallInit( struct sRemoteCallPort *psPort ) {
	if ( !g_thread_supported() ) {
		g_thread_init( NULL );
	}

#ifdef G_OS_WIN32
	if ( _pipe( psPort -> anFd, 2, O_BINARY ) == -1 ) {
#else
	if ( pipe( psPort -> anFd ) < 0 ) {
#endif
		Debug( KERN_DEBUG, "Failed to create pipe\n" );
		return -1;
	}

#ifdef G_OS_WIN32
	psPort -> psChannelIn = g_io_channel_win32_new_fd( psPort -> anFd[ 0 ] );
	psPort -> psChannelOut = g_io_channel_win32_new_fd( psPort -> anFd[ 1 ] );
#else
	psPort -> psChannelIn = g_io_channel_unix_new( psPort -> anFd[ 0 ] );
	psPort -> psChannelOut = g_io_channel_unix_new( psPort -> anFd[ 1 ] );
#endif
	g_io_channel_set_encoding( psPort -> psChannelIn, NULL, NULL );
	g_io_channel_set_encoding( psPort -> psChannelOut, NULL, NULL );

	psPort -> psCondition = g_cond_new();
	psPort -> psMutex = g_mutex_new();
	psPort -> nGtkInputTag = 0;

	if ( !psPort -> psCondition ) {
		if ( psPort -> psCondition ) {
			g_cond_free( psPort -> psCondition );
		}

		Debug( KERN_WARNING, "Could not create condition for remote call port\n" );
		return -1;
	}

	psPort -> psOwner = g_thread_self();

	return 0;
}

/**
 * \brief Register a new remote call port
 * \param psPort remote call port
 * \return error code (0 on success, otherwise error)
 */
int remoteCallRegister( struct sRemoteCallPort *psPort ) {
	if ( psPort -> nGtkInputTag != 0 ) {
		return 0;
	}

	psPort -> nGtkInputTag = g_io_add_watch_full( psPort -> psChannelIn, G_PRIORITY_HIGH, G_IO_IN, handleRemoteCall, psPort, NULL );
	if ( psPort -> nGtkInputTag != 0 ) {
		return 0;
	}

	Debug( KERN_WARNING, "Could not register remote call port\n" );
	return -1;
}

/**
 * \brief Remote call invoke
 * \param psPort remote call port
 * \param pFunc remote call function
 * \param pContext context pointer
 * \param pData user data
 * \return error code (0 on success, otherwise error)
 */
int remoteCallInvoke( struct sRemoteCallPort *psPort, remoteCallFunc pFunc, void *pContext, void *pData ) {
	struct sRemoteCall sEvent;
	gsize nWritten;
	GError *psError = NULL;
	GIOStatus nStatus;
	gchar *pnBuf;

	if ( psPort == NULL ) {
		Debug( KERN_WARNING, "psPort == NULL\n" );
		return -1;
	}

	if ( g_thread_self() == psPort -> psOwner ) {
		pFunc( pContext, pData );
		return 0;
	}

	sEvent.pFunc = pFunc;
	sEvent.psPort = psPort;
	sEvent.pContext = pContext;
	sEvent.pData = pData;

	pnBuf = ( gchar * ) &sEvent;

	nStatus = g_io_channel_write_chars( psPort -> psChannelOut, pnBuf, sizeof( sEvent ), &nWritten, &psError );
	if ( psError ) {
		Debug( KERN_DEBUG, "Error: %s\n", psError -> message );
		return -1;
	}

	if ( nStatus == G_IO_STATUS_NORMAL ) {
		if ( psPort -> psCondition ) {
			g_mutex_lock( psPort -> psMutex );
			g_io_channel_flush( psPort -> psChannelOut, NULL );
			g_cond_wait( psPort -> psCondition, psPort -> psMutex );
			g_mutex_unlock( psPort -> psMutex );
		} else {
			g_io_channel_flush( psPort -> psChannelOut, NULL );
		}

		return 0;
	}

	Debug( KERN_DEBUG, "Failed...\n" );

	return -1;
}

/**
 * \brief Get main remote port
 * \return main port
 */
struct sRemoteCallPort *remoteGetMainPort( void ) {
	return psMainPort;
}

/**
 * \brief Initialize remote port
 * \return 0
 */
int InitRemotePort( void ) {
	psMainPort = g_malloc0( sizeof( struct sRemoteCallPort ) );

	remoteCallInit( psMainPort );
	remoteCallRegister( psMainPort );

	return 0;
}
