/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Jan-Michael Brummer                                  *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file server.c
 * \brief Server handling
 */

#include <ffgtk.h>
#include <thread.h>

#define BUFFER_LENGTH 1024

#include <remote.h>

/**
 * \brief Remote call of processFax()
 * \param pContext context pointer, containing file name
 * \param pData data pointer
 */
void remoteSendFax( void *pContext, void *pData ) {
	processFax( pContext );
}

/**
 * \brief Print server thread
 * \param pData server socket
 * \return NULL
 */
gpointer printServerThread( gpointer pData ) {
	GSocket *psServer = pData;
	GSocket *psSock;
	GError *psError = NULL;
	gsize nLen;
	gchar *pnFileName;
	char anBuffer[ BUFFER_LENGTH ];
	ssize_t nWriteResult;
	ssize_t nWritten;

	while ( TRUE ) {
		psSock = g_socket_accept( psServer, NULL, &psError );
		g_assert_no_error( psError );

		pnFileName = g_strdup_printf( "%s/fax-XXXXXX", g_get_tmp_dir() );
		int nFileId = g_mkstemp( pnFileName );

		if ( nFileId == -1 ) {
			Debug( KERN_WARNING, _( "Can't open temporary file\n" ) );
			continue;
		}

		Debug( KERN_DEBUG, "file: %s (%d)\n", pnFileName, nFileId );

		do {
			nLen = g_socket_receive( psSock, anBuffer, BUFFER_LENGTH, NULL, &psError );

			if ( nLen > 0 ) {
				nWritten = 0;
				do {
					nWriteResult = write( nFileId, anBuffer + nWritten, nLen );
					if ( nWriteResult > 0 ) {
						nWritten += nWriteResult;
					}
				} while ( nLen != nWritten && ( nWriteResult != -1 || errno == EINTR ) );
			}
		} while ( nLen > 0 );

		if ( nLen == 0 ) {
			Debug( KERN_DEBUG, _( "Print job received on socket\n" ) );

			remoteCallInvoke( remoteGetMainPort(), remoteSendFax, pnFileName, NULL );
		}

		g_socket_close( psSock, &psError );
		g_assert_no_error( psError );
	}

	return NULL;
}

/**
 * \brief Initialize print server thread
 */
void PrintServerInit( void ) {
	GSocket *psSocket = NULL;
	GInetAddress *psInetAddress = NULL;
	GSocketAddress *psSockAddress = NULL;
	GError *psError = NULL;
	struct sThread sPrintThread;

	psSocket = g_socket_new( G_SOCKET_FAMILY_IPV4, G_SOCKET_TYPE_STREAM, G_SOCKET_PROTOCOL_TCP, &psError );
	if ( psError != NULL ) {
		Debug( KERN_DEBUG, "Could not create socket. Error: '%s'\n", psError -> message );
		g_error_free( psError );
		return;
	}

	psInetAddress = g_inet_address_new_from_string( "127.0.0.1" );

	psSockAddress = g_inet_socket_address_new( psInetAddress, 9901 );
	if ( psSockAddress == NULL ) {
		Debug( KERN_DEBUG, "Could not create sock address on port 9901\n" );
		g_object_unref( psSocket );
		return;
	}

	psError = NULL;
	if ( g_socket_bind( psSocket, psSockAddress, TRUE, &psError ) == FALSE ) {
		Debug( KERN_DEBUG, "Could not bind to socket. Error: %s\n", psError -> message );
		g_error_free( psError );
		g_object_unref( psSocket );
		return;
	}

	if ( g_socket_listen( psSocket, &psError ) == FALSE ) {
		Debug( KERN_DEBUG, "Could not listen on socket. Error: %s\n", psError -> message );
		g_error_free( psError );
		g_object_unref( psSocket );
		return;
	}

	Debug( KERN_DEBUG, _( "Fax Server running on port 9901\n" ) );

	InitThread( &sPrintThread, "printserver" );
	StartThread( &sPrintThread, printServerThread, psSocket );
}
