/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Jan-Michael Brummer                                  *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file trayicon.c
 * \brief Trayicon related functions
 */

#include <ffgtk.h>
#include <menucallbacks.h>
#include <trayicon.h>
#include <osdep/osdep.h>

#ifdef HAVE_LIBNOTIFY
#include <notification.h>
#endif

/** Widget containing trayicon */
static GtkStatusIcon *psTrayIcon = NULL;

/** is menu visible */
gboolean bMenuVisible = FALSE;

/** main gtk loop */
extern GMainLoop *psLoop;
static GtkWidget *psMenu = NULL;

#ifdef G_OS_WIN32
/* This is a workaround for a bug in windows GTK+. Clicking outside of the
 * menu does not get rid of it, so instead we get rid of it as soon as the
 * pointer leaves the menu.
 */

static gboolean hide_docklet_menu( gpointer data ) {
	if (data != NULL) {
		gtk_menu_popdown( GTK_MENU( data ) );
	}

	return FALSE;
}

gboolean trayIconLeaveEnter( GtkWidget *psMenuWidget, GdkEventCrossing *psEvent, void *pData ) {
	static guint nHideTimer = 0;

	if ( psEvent -> type == GDK_LEAVE_NOTIFY && ( psEvent -> detail == GDK_NOTIFY_ANCESTOR || psEvent -> detail == GDK_NOTIFY_UNKNOWN ) ) {
		if ( nHideTimer == 0 ) {
			nHideTimer = g_timeout_add( 500, hide_docklet_menu, psMenuWidget );
		}
	} else if ( psEvent -> type == GDK_ENTER_NOTIFY && psEvent -> detail == GDK_NOTIFY_ANCESTOR ) {
		if ( nHideTimer != 0 ) {
			g_source_remove( nHideTimer );
			nHideTimer = 0;
		}
	}

	return FALSE;
}
#endif

/**
 * \brief Handle tray icon menu
 * \param psStatusIcon status icon pointer
 * \param nButton which button
 * \param nActivationTime activation time
 * \param pUserData user data pointer
 */
static void trayIconOnMenu( GtkStatusIcon *psStatusIcon, guint nButton, guint nActivateTime, gpointer pUserData) {
	if ( psMenu != NULL ) {
		gtk_widget_destroy( psMenu );
		psMenu = NULL;
	}
	psMenu = createMenu( TRUE );

#ifdef G_OS_WIN32
	/* Workaround for several Windows bugs */
	g_signal_connect( psMenu, "leave-notify-event", G_CALLBACK( trayIconLeaveEnter ), NULL );
	g_signal_connect( psMenu, "enter-notify-event", G_CALLBACK( trayIconLeaveEnter ), NULL );
#endif /* G_OS_WIN32 */
	gtk_widget_show_all( psMenu );

#ifdef G_OS_WIN32
	gtk_menu_popup( GTK_MENU( psMenu ), NULL, NULL, gtk_status_icon_position_menu, psStatusIcon, 0, nActivateTime );
#else
	gtk_menu_popup( GTK_MENU( psMenu ), NULL, NULL, gtk_status_icon_position_menu, psStatusIcon, nButton, nActivateTime );
#endif /* G_OS_WIN32 */
}

/**
 * \brief Create tray icon and set menu/tooltip
 */
static void createTrayIcon( void ) {
	psTrayIcon = gtk_status_icon_new();
	g_signal_connect( G_OBJECT( psTrayIcon ), "popup-menu", G_CALLBACK( trayIconOnMenu ), NULL );
	g_signal_connect( G_OBJECT( psTrayIcon ), "activate", G_CALLBACK( trayIconToggled ), NULL );
	trayIconSetIconType( TYPE_CALLOUT );
	gtk_status_icon_set_tooltip_text( psTrayIcon,  "ffgtk" );
	gtk_status_icon_set_visible( psTrayIcon, TRUE );
}

/**
 * \brief Periodically update tooltip information
 * \param pUserData pointer to tray icon
 * \return TRUE
 */
static gboolean updateTooltip( gpointer pUserData ) {
	gchar *pnIp = NULL;
	gchar *pnText = NULL;
	gint nMaxUp = 0;
	gint nMaxDown = 0;

	pnIp = routerGetIp( getActiveProfile() );

	if ( pnIp != NULL ) {
		routerGetSpeed( getActiveProfile(), &nMaxUp, &nMaxDown );
		pnText = g_strdup_printf( _( "IP: %s\nMax. Up: %d kbit/s\nMax. Down: %d kbit/s" ), pnIp, nMaxUp >> 10, nMaxDown >> 10 );
		g_free( pnIp );
		pnIp = NULL;
	} else {
		pnText = g_strdup_printf( _( "Not connected" ) );
	}

	GtkStatusIcon *psIcon = ( GtkStatusIcon * ) pUserData;
	gtk_status_icon_set_tooltip_text( psIcon, pnText );

	g_free( pnText );
	pnText = NULL;

	return TRUE;
}

/**
 * \brief Get tray icon widget
 * \return trayicon widget
 */
GtkStatusIcon *getTrayIcon( void ) {
	return psTrayIcon;
}

/**
 * \brief Set which tray icon to use 
 * \input type if icon    
 */

void trayIconSetIconType( int nType ) {
	Debug( KERN_DEBUG, "Trayicon type set to %d\n", nType );
	gtk_status_icon_set_from_pixbuf( GTK_STATUS_ICON( getTrayIcon() ), getTypeIcon( nType ) );
}

/**
 * \brief Recreate app menu
 */
void trayIconRecreateAppMenu( void ) {
}

/**
 * \brief is any tray/dock menu system available?
 * \return tray menu available
 */

gboolean menuAvailable( void ) {
	if (psTrayIcon != NULL ) {
		return bMenuVisible;
	}
	return FALSE;
}

/**
 * \brief Initialize tray icon structure and start update tooltip thread
 */
void trayIconInit( void ) {
	createTrayIcon();
	g_timeout_add_seconds( 15, updateTooltip, psTrayIcon );

#ifdef HAVE_LIBNOTIFY
	if ( !notificationNotifyHasPersistance() ) {
		/* Notification server has persistance, so tray icon is most likely not visible */
		bMenuVisible = FALSE;
	}
#endif
}
