/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Louis Lagendijk                                      *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file sources-registry.c
 * \brief Interface for the sources registry
 */
#include <ffgtk.h>
#include <libebook/libebook.h>
#include <ab_ebook/ebook-sources.h>
#include <addressbook.h>

/**
 * \brief Retrieve source registry
 * \return Registry pointer
 */
static ESourceRegistry *getSourceRegistry( void ) {
	static ESourceRegistry *psRegistry = NULL;
	GError *psError = NULL;
	if( psRegistry == NULL ) {
		psRegistry = e_source_registry_new_sync( NULL, &psError );
	}
	return psRegistry;
}

/**
 * \brief Get ESource for active address book
 * \return Esource
 */
static ESource  *getSelectedEbookESource( void ) {
        ESource *psSource = NULL;
        const gchar *psUidString = getSelectedEbookId();
        psSource = e_source_registry_ref_source( getSourceRegistry(), psUidString );
	return psSource;
}

/**
 * \brief Retrieve the uid for the source
 * \param The source
 * \return the uid of the source
 */
const gchar *getEbookId( ESource *psEsource ) {
	return e_source_get_uid( psEsource );
}

/**
 * \brief Retrieve the display name for the source
 * \param The source
 * \return the display name of the source
 */
const gchar *getEbookName( ESource *psEsource ) {
	return e_source_get_display_name( psEsource );
}

/**
 * \brief Get selected evolution address book Id
 * \return evolution address book
 */
const gchar *getSelectedEbookId( void ) {
        return prefsGetString( getActiveProfile(), "/plugins/evolution/book" );
}

/**
 * \brief Set selected evolution address book
 * \param pnUri evolution address book
 */
void setSelectedEbookId( const gchar *pnId ) {
	prefsSetString( getActiveProfile(), "/plugins/evolution/book", pnId );
}

/**
 * \brief Get client for active address book
 * \return ebook client
 */
EBookClient  *getSelectedEbookClient( void ) {
	EBookClient *psClient;
	GError *psError = NULL;
        ESource *psSource = NULL;
        psSource = getSelectedEbookESource();
        psClient = e_book_client_new( psSource,  &psError );
	return psClient;
}

/**
 * \brief Retrieve a list of all address book sources
 * \return List of address book sources
 */
GList *getEbooksList( void ) {

	GList *psSource, *psSources;
	GList *psEbooksList = NULL;
	struct sEbookData *psEbookData = NULL;
	psSources = e_source_registry_list_sources ( getSourceRegistry(), E_SOURCE_EXTENSION_ADDRESS_BOOK );

	for ( psSource = psSources; psSource != NULL; psSource = psSource -> next ) {
		ESource *psESource = E_SOURCE( psSource -> data );
		psEbookData = g_new( struct sEbookData, 1);
		psEbookData -> pnEbookName = g_strdup( getEbookName( psESource  ) );
		psEbookData -> pnEbookId = g_strdup( getEbookId ( psESource ) );

		psEbooksList = g_list_append( psEbooksList, psEbookData );
	}
	g_list_free_full( psSources, g_object_unref );
	return psEbooksList;
}

/**
 * \brief free data allocated for an address book data item
 * \param ebook data item to be destroyed
 */
static void freeEbookData( struct sEbookData *psEbookData ) {
	g_free( psEbookData -> pnEbookName );
	g_free( psEbookData -> pnEbookId );
}

/**
 * \brief free ebook list
 */
void freeEbooksList( GList *psEbooksList ) {
	g_list_free_full( psEbooksList, (GDestroyNotify) freeEbookData );
}	
