/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Jan-Michael Brummer                                  *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file ab_fritzfon.c
 * \brief FritzFon addressbook plugin
 */

#include <ffgtk.h>
#include <sys/stat.h>

extern const gchar *getProfilesDir( void );

static xmlnode *psMasterNode = NULL;
static struct sAddressBook sFritzFon;

#define VALID_NUMBER( x ) ( ( x != NULL ) && ( strlen( x ) > 0 ) )

/** Swap first/last name */
static gboolean fritzfonGetSwap( void ) {
	return prefsGetBool( getActiveProfile(), "/plugins/fritzfon/swap" );
}

/** Swap first/last name */
static void fritzfonSetSwap( gboolean bSwap ) {
	prefsSetBool( getActiveProfile(), "/plugins/fritzfon/swap", bSwap );
}

/**
 * \brief Get profile book file name
 * \return file name
 */
static gchar *getBookFile( void ) {
	gchar *pnFile;
	
	pnFile = g_build_filename( getProfilesDir(), getActiveProfile() -> pnName, "fritzfon.xml", NULL );

	return pnFile;
}

static xmlnode *createPhone( char *pnType, char *pnNumber ) {
	xmlnode *psNumber;

	psNumber = xmlnode_new( "number" );
	xmlnode_set_attrib( psNumber, "type", pnType );

	/* For the meantime set priority to 0, TODO */
	xmlnode_set_attrib( psNumber, "prio", "0" );
	xmlnode_insert_data( psNumber, pnNumber, -1 );

	return psNumber;
}

/**
 * \brief Convert person structure to xml node
 * \param psPerson person structure
 * \return xml node
 */
static xmlnode *contactToXmlnode( struct sPerson *psPerson ) {
	xmlnode *psNode;
	xmlnode *psPersonNode;
	xmlnode *psRealName;
	xmlnode *psImage;
	xmlnode *psTelephony;
	xmlnode *psCategory;
	gchar *pnTmp = NULL;

	/* Main contact entry */
	psNode = xmlnode_new( "contact" );

	/* Category */
	psCategory = xmlnode_new( "category" );
	xmlnode_insert_data( psCategory, psPerson -> pnCategory, -1 );
	xmlnode_insert_child( psNode, psCategory );

	/* Person */
	psPersonNode = xmlnode_new( "person" );

	/* RealName */
	if ( fritzfonGetSwap() == FALSE ) {
		pnTmp = g_strdup_printf( "%s %s", psPerson -> pnFirstName, psPerson -> pnLastName );
	} else {
		pnTmp = g_strdup_printf( "%s %s", psPerson -> pnLastName, psPerson -> pnFirstName );
	}

	psRealName = xmlnode_new( "realName" );
	xmlnode_insert_data( psRealName, pnTmp, -1 );
	xmlnode_insert_child( psPersonNode, psRealName );
	g_free( pnTmp );

	/* ImageURL stub */
	psImage = xmlnode_new( "ImageURL" );
	xmlnode_insert_child( psPersonNode, psImage );

	/* Insert person to main node */
	xmlnode_insert_child( psNode, psPersonNode );

	/* Telephony */
	psTelephony = xmlnode_new( "telephony" );

	/* Add home phone */
	if ( psPerson -> pnPrivatePhone != NULL && strlen( psPerson -> pnPrivatePhone ) > 0 ) {
		xmlnode *psNumber;

		psNumber = xmlnode_new( "number" );
		xmlnode_set_attrib( psNumber, "type", "home" );

		/* For the meantime set priority to 0, TODO */
		xmlnode_set_attrib( psNumber, "prio", "0" );
	
		xmlnode_insert_data( psNumber, psPerson -> pnPrivatePhone, -1 );
		xmlnode_insert_child( psTelephony, psNumber );
	}

	/* Add mobile phone */
	if ( psPerson -> pnPrivateMobile != NULL && strlen( psPerson -> pnPrivateMobile ) > 0 ) {
		xmlnode *psNumber;

		psNumber = xmlnode_new( "number" );
		xmlnode_set_attrib( psNumber, "type", "mobile" );

		/* For the meantime set priority to 0, TODO */
		xmlnode_set_attrib( psNumber, "prio", "0" );
		xmlnode_insert_data( psNumber, psPerson -> pnPrivateMobile, -1 );
		xmlnode_insert_child( psTelephony, psNumber );
	}

	/* Add work phone */
	if ( psPerson -> pnBusinessPhone != NULL && strlen( psPerson -> pnBusinessPhone ) > 0 ) {
		xmlnode *psNumber;

		psNumber = xmlnode_new( "number" );
		xmlnode_set_attrib( psNumber, "type", "work" );

		/* For the meantime set priority to 0, TODO */
		xmlnode_set_attrib( psNumber, "prio", "0" );
		xmlnode_insert_data( psNumber, psPerson -> pnBusinessPhone, -1 );
		xmlnode_insert_child( psTelephony, psNumber );
	}

	/* Add work fax */
	if ( psPerson -> pnBusinessFax != NULL && strlen( psPerson -> pnBusinessFax ) > 0 ) {
		xmlnode *psNumber;

		psNumber = xmlnode_new( "number" );
		xmlnode_set_attrib( psNumber, "type", "fax_work" );

		/* For the meantime set priority to 0, TODO */
		xmlnode_set_attrib( psNumber, "prio", "0" );
		xmlnode_insert_data( psNumber, psPerson -> pnBusinessFax, -1 );
		xmlnode_insert_child( psTelephony, psNumber );
	}

	xmlnode_insert_child( psNode, psTelephony );

	return psNode;
}

/**
 * \brief Convert phonebooks to xml node
 * \return xml node
 */
xmlnode *phonebooksToXmlnode( void ) {
	xmlnode *psNode;
	xmlnode *psChild;
	xmlnode *psBook;
	GList *psList = psPersonsList;

	/* Create general phonebooks node */
	psNode = xmlnode_new( "phonebooks" );

	/* Currently we only support one phonebook, TODO */
	psBook = xmlnode_new( "phonebook" );
	xmlnode_set_attrib( psBook, "name", "Telefonbuch" );
	xmlnode_insert_child( psNode, psBook );

	/* Loop through persons list and add only non-deleted entries */
	for ( psList = psPersonsList; psList != NULL && psList -> data != NULL; psList = psList -> next ) {
		struct sPerson *psPerson = psList -> data;

		/* Skip deleted persons */
		if ( psPerson -> nFlags & PERSON_FLAG_DELETED ) {
			psPerson -> nFlags = PERSON_FLAG_UNCHANGED;
			continue;
		}

		if ( !VALID_NUMBER( psPerson -> pnPrivatePhone ) &&
			!VALID_NUMBER( psPerson -> pnPrivateMobile ) &&
			!VALID_NUMBER( psPerson -> pnBusinessPhone ) &&
			!VALID_NUMBER( psPerson -> pnBusinessFax ) ) {
			continue;
		}

		/* Convert each contact and add it to current phone book */
		psChild = contactToXmlnode( psPerson );
		xmlnode_insert_child( psBook, psChild );
	}

	return psNode;
}

/**
 * \brief Add contact to persons list
 * \param psNode xml contact node
 * \param nCount unique person id
 */
static void addContact( xmlnode *psNode, int nCount ) {
	xmlnode *psPerson;
	xmlnode *psRealName;
	xmlnode *psImage;
	xmlnode *psTelephony;
	xmlnode *psChild;
	xmlnode *psCategory;
	gchar *pnFirstName;
	gchar *pnLastName;
	gchar *pnHome = NULL;
	gchar *pnWork = NULL;
	gchar *pnMobile = NULL;
	gchar *pnFaxWork = NULL;
	gchar *pnTmp;
	gchar *pnRealName;
	gchar *pnImage;
	gchar *pnCategory = NULL;
	GdkPixbuf *psImageBuf = NULL;

	/* Get person entry */
	psPerson = xmlnode_get_child( psNode, "person" );
	if ( psPerson == NULL ) {
		return;
	}

	/* Get real name entry */
	psRealName = xmlnode_get_child( psPerson, "realName" );
	if ( psRealName == NULL ) {
		return;
	}	
	pnRealName = xmlnode_get_data( psRealName );

	pnTmp = strchr( pnRealName, ' ' );
	if ( pnTmp != NULL ) {
		gint nLen = 0;
		pnLastName = strrchr( pnRealName, ' ' ) + 1;
		nLen = strlen( pnRealName ) - strlen( pnLastName ) - 1;
		pnFirstName = g_malloc0( nLen + 1 );
		strncpy( pnFirstName, pnRealName, nLen );
		pnFirstName[ nLen ] = '\0';
	} else {
		pnFirstName = NULL;
		pnLastName = pnRealName;
	}

	psTelephony = xmlnode_get_child( psNode, "telephony" );
	if ( psTelephony != NULL ) {
		/* Check for numbers */
		for ( psChild = xmlnode_get_child( psTelephony, "number" ); psChild != NULL; psChild = xmlnode_get_next_twin( psChild ) ) {
			const gchar *pnType;

			pnType = xmlnode_get_attrib( psChild, "type" );
			if ( pnType == NULL ) {
				continue;
			}

			if ( strcmp( pnType, "mobile" ) == 0 ) {
				pnMobile = xmlnode_get_data( psChild );
			} else if ( strcmp( pnType, "home" ) == 0 ) {
				pnHome = xmlnode_get_data( psChild );
			} else if ( strcmp( pnType, "work" ) == 0 ) {
				pnWork = xmlnode_get_data( psChild );
			} else if ( strcmp( pnType, "fax_work" ) == 0 ) {
				pnFaxWork = xmlnode_get_data( psChild );
			}
		}
	}

	/* Get image */
	psImage = xmlnode_get_child( psPerson, "imageURL" );
	if ( psImage != NULL ) {
		pnImage = xmlnode_get_data( psImage );
		if ( pnImage != NULL ) {
			/* file:///var/InternerSpeicher/FRITZ/fonpix/946684999-0.jpg */
			if ( strlen( pnImage ) > 28 ) {
				struct sUrlHandler *psHandler = NULL;
				const gchar *pnUser = NULL;
				const gchar *pnPassword = NULL;
				gchar *pnUsrPwd = NULL;
				struct sProfile *psProfile = getActiveProfile();
				int nPos = findString( pnImage, 0, "/ftp/" );
				//nPos = findStringReverse( pnImage, nPos, "/" );

				if ( nPos == -1 ) {
					nPos = findString( pnImage, 0, "/FRITZ/" );
				} else {
					nPos += 5;
				}

				Debug( KERN_DEBUG, "pnImage: %s\n", pnImage );
				gchar *pnFile = g_strdup_printf( "ftp://%s/%s", routerGetHost( getActiveProfile() ), pnImage + nPos );
				int nError;
				int nResult;

				Debug( KERN_DEBUG, "pnFile: %s\n", pnFile );
				pnUser = voiceBoxGetUser( psProfile );
				pnPassword = voiceBoxGetPassword( psProfile );
				if ( pnUser != NULL && strlen( pnUser ) > 0 ) {
					pnUsrPwd = g_strdup_printf( "%s:%s", pnUser, pnPassword );
				}

				psHandler = urlHandler( pnFile, 21 );

				if ( pnUsrPwd != NULL ) {
					curl_easy_setopt( psHandler -> psHandle, CURLOPT_USERPWD, pnUsrPwd );
					Debug( KERN_DEBUG, "Set user/pwd\n" );
				}
				nError = readUrl( psHandler, psProfile );
				Debug( KERN_DEBUG, "nError: %d\n", nError );
				if ( nError == 0 ) {
					char *pnTmpFile = g_strdup_printf( "%s/test.jpg", g_get_tmp_dir() );
					FILE *psFile = fopen( pnTmpFile, "wb+" );

					if ( psFile != NULL ) {
						nResult = fwrite( psHandler -> pnData, 1, psHandler -> nSize, psFile );
						if ( nResult != psHandler -> nSize ) {
							Debug( KERN_DEBUG, "Could not save image\n" );
						}
						fclose( psFile );
						psImageBuf = gdk_pixbuf_new_from_file( pnTmpFile, NULL );
					}
					freeHandler( psHandler );

					g_free( pnTmpFile );
				}
			}
		}
	}

	psCategory = xmlnode_get_child( psNode, "category" );
	if ( psCategory != NULL ) {
		pnCategory = xmlnode_get_data( psCategory );
	}

	/* Create person hash table, fill structure and add them to internal list */
	GHashTable *psTable = g_hash_table_new( NULL, NULL );
	gchar *pnId = g_strdup_printf( "%d", nCount );

	AddInfo( psTable, PERSON_ID, pnId );
	if ( fritzfonGetSwap() == FALSE ) {
		AddInfo( psTable, PERSON_FIRST_NAME, pnFirstName );
		AddInfo( psTable, PERSON_LAST_NAME, pnLastName );
	} else {
		AddInfo( psTable, PERSON_FIRST_NAME, pnLastName );
		AddInfo( psTable, PERSON_LAST_NAME, pnFirstName );
	}
	AddInfo( psTable, PERSON_DISPLAY_NAME, pnRealName );
	AddInfo( psTable, PERSON_BUSINESS_PHONE, pnWork );
	AddInfo( psTable, PERSON_PRIVATE_PHONE, pnHome );
	AddInfo( psTable, PERSON_PRIVATE_MOBILE, pnMobile );
	AddInfo( psTable, PERSON_BUSINESS_FAX, pnFaxWork );
	AddInfo( psTable, PERSON_IMAGE, ( const gchar * ) psImageBuf );
	AddInfo( psTable, PERSON_CATEGORY, pnCategory );

	AddPerson( psTable, FALSE );
	g_free( pnId );

	g_hash_table_destroy( psTable );
}

/**
 * \brief Add phonebook
 * \param psNode phonebook xml node
 */
static void addPhonebook( xmlnode *psNode ) {
	xmlnode *psChild;
	int nCount = 0;

	for ( psChild = xmlnode_get_child( psNode, "contact" ); psChild != NULL; psChild = xmlnode_get_next_twin( psChild ) ) {
		addContact( psChild, nCount++ );
	}
}

void fritzfonTypeDetection( void ) {
	struct sProfile *psProfile = getActiveProfile();

	if ( psProfile != NULL ) {
		if ( strstr( psProfile -> sRouterInfo.pnFirmware, ".05." ) ) {
			sFritzFon.nReadFlags |= PERSON_BUSINESS_FAX;
			sFritzFon.nWriteFlags |= PERSON_BUSINESS_FAX;
		}
	}
}

/**
 * \brief Read ebook list
 * \return error code
 */
static int fritzfonReadBook( void ) {
	struct sUrlHandler *psHandler;
	gchar anBuffer[ 1024 ];
	gint nError = -1;
	xmlnode *psNode = NULL;
	xmlnode *psChild;
	struct curl_httppost* post = NULL;
	struct curl_httppost* last = NULL;
	struct curl_slist *m_headerlist = NULL;

	fritzfonTypeDetection();

	if ( routerLogin( getActiveProfile() ) == -1 ) {
		return -1;
	}

	snprintf( anBuffer, sizeof( anBuffer ), "%s/cgi-bin/firmwarecfg", routerGetHost( getActiveProfile() ) );
	psHandler = urlHandler( anBuffer, 80 );

	m_headerlist = curl_slist_append(m_headerlist, "ENCTYPE=\"multipart/form-data\""); 	
	curl_easy_setopt(psHandler -> psHandle, CURLOPT_HTTPHEADER, m_headerlist);

	curl_formadd(&post, &last, CURLFORM_COPYNAME, "sid",   CURLFORM_COPYCONTENTS, getActiveProfile() -> sRouterInfo.pnSessionId, CURLFORM_END); 
	curl_formadd(&post, &last, CURLFORM_COPYNAME, "PhonebookId",   CURLFORM_COPYCONTENTS, "0", CURLFORM_END); 
	curl_formadd(&post, &last, CURLFORM_COPYNAME, "PhonebookExportName",   CURLFORM_COPYCONTENTS, "Telefonbuch", CURLFORM_END); 
	curl_formadd(&post, &last, CURLFORM_COPYNAME, "PhonebookExport",   CURLFORM_COPYCONTENTS, "", CURLFORM_END); 

	curl_easy_setopt(psHandler -> psHandle, CURLOPT_HTTPPOST, post); 

	/* set timeout to 3sec */
	//curl_easy_setopt( psHandler -> psHandle, CURLOPT_TIMEOUT, 3 );

	nError = readUrl( psHandler, getActiveProfile() );
	if ( nError == 0 ) {
		gchar *pnFile;

		pnFile = getBookFile();
		saveData( pnFile, psHandler -> pnData, psHandler -> nSize );

		psNode = readXmlFromFile( pnFile, "phonebooks" );
		if ( psNode == NULL ) {
			Debug( KERN_DEBUG, "Could not read %s\n", pnFile );
			g_free( pnFile );
			freeHandler( psHandler );
			return -1;
		}
		g_free( pnFile );

		psMasterNode = psNode;

		for ( psChild = xmlnode_get_child( psNode, "phonebook" ); psChild != NULL; psChild = xmlnode_get_next_twin( psChild ) ) {
			addPhonebook( psChild );
		}
	}
	freeHandler( psHandler );

	routerLogout( getActiveProfile() );

	return 0;
}

xmlnode *fritzfonFindPersonEntry( gchar *pnName ) {
	xmlnode *psNode = xmlnode_get_child( psMasterNode, "phonebook");
	xmlnode *psChild = NULL;

	if ( psNode == NULL ) {
		return NULL;
	}

	for ( psChild = xmlnode_get_child( psNode, "contact" ); psChild != NULL; psChild = xmlnode_get_next_twin( psChild ) ) {
		xmlnode *psEntry;
		gchar *pnEntryName;

		psEntry = xmlnode_get_child( psChild, "person/realName" );
		if ( psEntry != NULL ) {
			pnEntryName = xmlnode_get_data( psEntry );
			if ( pnEntryName ) {
				Debug( KERN_DEBUG, "pnEntryName: '%s'\n", pnEntryName );
			}
			if ( pnEntryName && pnName && !strcmp( pnEntryName, pnName ) ) {
				return psEntry -> psParent -> psParent;
			}
		}
	}

	return NULL;
}

int xmlnode_set_data( xmlnode *, gchar * );

/**
 * \brief Write ebook list
 * \return error code
 */
static int fritzfonWriteBook( void ) {
	struct sUrlHandler *psHandler;
	gchar anBuffer[ 1024 ];
	gint nError = -1;
	//xmlnode *node;
	char *pnData;

	fritzfonTypeDetection();

	{
		int nLen;
		xmlnode *psNode = psMasterNode;
		GList *psList = psPersonsList;
		xmlnode *psBook = NULL;

		/* In case we do not have a valid masternode, create a new master node */
		if ( psNode == NULL ) {
			psNode = xmlnode_new( "phonebooks" );
			xmlnode_new_child( psNode, "phonebook" );
		}
		psBook = xmlnode_get_child( psNode, "phonebook" );

		/* Loop through persons list and add only non-deleted entries */
		for ( psList = psPersonsList; psList != NULL && psList -> data != NULL; psList = psList -> next ) {
			struct sPerson *psPerson = psList -> data;
			xmlnode *psEntry;

			/* Deleted persons */
			if ( psPerson -> nFlags & PERSON_FLAG_DELETED ) {
				psEntry = fritzfonFindPersonEntry( psPerson -> pnDisplayName );
				if ( psEntry != NULL ) {
					Debug( KERN_DEBUG, "Removing entry: '%s'\n", psPerson -> pnDisplayName );
					xmlnode_free( psEntry );
				}

				psPerson -> nFlags = PERSON_FLAG_UNCHANGED;
				continue;
			}
			/* Changed entry */
			if ( psPerson -> nFlags & PERSON_FLAG_CHANGED ) {
				psEntry = fritzfonFindPersonEntry( psPerson -> pnDisplayName );
				if ( psEntry == NULL ) {
					psPerson -> nFlags |= PERSON_FLAG_NEW;
				} else {
					xmlnode *psRealName;
					xmlnode *psTelephony;
					xmlnode *psChild;
					gchar *pnName;
					gboolean bMobileSet = false;
					gboolean bPrivateSet = false;
					gboolean bBusinessSet = false;
					gboolean bBusinessFaxSet = false;

					Debug( KERN_DEBUG, "Applying changes to: '%s'\n", psPerson -> pnDisplayName );

					/* Set real name entry */
					psRealName = xmlnode_get_child( psEntry, "person/realName" );
					if ( psRealName == NULL ) {
						return -1;
					}
					Debug( KERN_DEBUG, "Current name was: '%s'\n", xmlnode_get_data( psRealName ) );
					pnName = g_strdup_printf( "%s %s", psPerson -> pnFirstName, psPerson -> pnLastName );
					xmlnode_set_data( psRealName, pnName );
					Debug( KERN_DEBUG, "Current name is: '%s'\n", xmlnode_get_data( psRealName ) );

					/* Update phone numbers */
					psTelephony = xmlnode_get_child( psEntry, "telephony" );
					if ( psTelephony == NULL ) {
						return -1;
					}

					/* If present, change value */
					for ( psChild = xmlnode_get_child( psTelephony, "number" ); psChild != NULL; psChild = xmlnode_get_next_twin( psChild ) ) {
						const gchar *pnType;

						pnType = xmlnode_get_attrib( psChild, "type" );
						if ( pnType == NULL ) {
							continue;
						}

						if ( strcmp( pnType, "mobile" ) == 0 ) {
							/* We have a mobile entry and we also have a mobile entry in our current database */
							if ( psPerson -> pnPrivateMobile != NULL && strlen( psPerson -> pnPrivateMobile ) > 0 ) {
								xmlnode_set_data( psChild, psPerson -> pnPrivateMobile );
							} else {
								/* free structure */
								xmlnode_set_data( psChild, "" );
							}
							bMobileSet = true;
						} else if ( strcmp( pnType, "home" ) == 0 ) {
							/* We have a home entry and we also have a home entry in our current database */
							if ( psPerson -> pnPrivatePhone != NULL && strlen( psPerson -> pnPrivatePhone ) > 0 ) {
								xmlnode_set_data( psChild, psPerson -> pnPrivatePhone );
							} else {
								/* free structure */
								xmlnode_set_data( psChild, "" );
							}
							bPrivateSet = true;
						} else if ( strcmp( pnType, "work" ) == 0 ) {
							/* We have a business entry and we also have a business entry in our current database */
							if ( psPerson -> pnBusinessPhone != NULL && strlen( psPerson -> pnBusinessPhone ) > 0 ) {
								xmlnode_set_data( psChild, psPerson -> pnBusinessPhone );
							} else {
								/* free structure */
								xmlnode_set_data( psChild, "" );
							}
							bBusinessSet = true;
						} else if ( strcmp( pnType, "fax_work" ) == 0 ) {
							/* We have a business fax entry and we also have a business fax entry in our current database */
							if ( psPerson -> pnBusinessFax != NULL && strlen( psPerson -> pnBusinessFax ) > 0 ) {
								xmlnode_set_data( psChild, psPerson -> pnBusinessFax );
							} else {
								/* free structure */
								xmlnode_set_data( psChild, "" );
							}
							bBusinessFaxSet = true;
						}
					}

					/* Otherwise add new number */
					if ( bMobileSet == false && psPerson -> pnPrivateMobile != NULL && strlen( psPerson -> pnPrivateMobile ) > 0 )  {
						xmlnode *psPhone = createPhone( "mobile", psPerson -> pnPrivateMobile );
						if ( psPhone != NULL ) {
							xmlnode_insert_child( psTelephony, psPhone );
						}
					}

					if ( bPrivateSet == false && psPerson -> pnPrivatePhone != NULL && strlen( psPerson -> pnPrivatePhone ) > 0 )  {
						xmlnode *psPhone = createPhone( "home", psPerson -> pnPrivatePhone );
						if ( psPhone != NULL ) {
							xmlnode_insert_child( psTelephony, psPhone );
						}
					}

					if ( bBusinessSet == false && psPerson -> pnBusinessPhone != NULL && strlen( psPerson -> pnBusinessPhone ) > 0 )  {
						xmlnode *psPhone = createPhone( "work", psPerson -> pnBusinessPhone );
						if ( psPhone != NULL ) {
							xmlnode_insert_child( psTelephony, psPhone );
						}
					}

					if ( bBusinessFaxSet == false && psPerson -> pnBusinessFax != NULL && strlen( psPerson -> pnBusinessFax ) > 0 )  {
						xmlnode *psPhone = createPhone( "fax_work", psPerson -> pnBusinessFax );
						if ( psPhone != NULL ) {
							xmlnode_insert_child( psTelephony, psPhone );
						}
					}
				}
			}

			/* New person */
			if ( psPerson -> nFlags & PERSON_FLAG_NEW ) {
				psEntry = fritzfonFindPersonEntry( psPerson -> pnDisplayName );
				if ( psEntry == NULL ) {
					Debug( KERN_DEBUG, "Adding new person: '%s'\n", psPerson -> pnDisplayName );
					xmlnode *psChild = contactToXmlnode( psPerson );
					xmlnode_insert_child( psBook, psChild );
					psPerson -> nFlags = PERSON_FLAG_UNCHANGED;
					continue;
				} else {
					Debug( KERN_WARNING, "Trying to add duplicate entry '%s'\n", psPerson -> pnDisplayName );
				}
			}
		}

		pnData = xmlnode_to_formatted_str( psNode, &nLen );
		printf( "nLen: %d\n", nLen );
		if ( nLen > 0 ) {
			saveDebugData( "test.xml", pnData, nLen );
		}
	}

	Debug( KERN_DEBUG, "Login in\n" );
	if ( routerLogin( getActiveProfile() ) == -1 ) {
		return -1;
	}

	if ( pnData != NULL ) {
		gchar *pnFile;

		pnFile = getBookFile();
		saveData( pnFile, pnData, -1 );

		struct curl_httppost* post = NULL;
		struct curl_httppost* last = NULL;
		struct curl_slist *m_headerlist = NULL;


		snprintf( anBuffer, sizeof( anBuffer ), "%s/cgi-bin/firmwarecfg", routerGetHost( getActiveProfile() ) );
		psHandler = urlHandler( anBuffer, 80 );

		m_headerlist = curl_slist_append(m_headerlist, "ENCTYPE=\"multipart/form-data\""); 	

		curl_easy_setopt(psHandler -> psHandle, CURLOPT_HTTPHEADER, m_headerlist);
		curl_formadd(&post, &last, CURLFORM_COPYNAME, "sid",   CURLFORM_COPYCONTENTS, getActiveProfile() -> sRouterInfo.pnSessionId, CURLFORM_END); 
		curl_formadd(&post, &last, CURLFORM_COPYNAME, "PhonebookId",   CURLFORM_COPYCONTENTS, "0", CURLFORM_END); 
		curl_formadd(&post, &last, CURLFORM_COPYNAME, "PhonebookImportFile", CURLFORM_FILE, pnFile, CURLFORM_END); 

		curl_easy_setopt(psHandler -> psHandle, CURLOPT_HTTPPOST, post); 

		Debug( KERN_DEBUG, "Write book\n" );
		nError = readUrl( psHandler, getActiveProfile() );
		freeHandler( psHandler );
		g_free(pnData);
		g_free( pnFile );
	}
	//xmlnode_free(node);

	Debug( KERN_DEBUG, "Log out\n" );
	routerLogout( getActiveProfile() );

	return nError;
}

void prefsAddNone( struct sProfile *psProfile, const char *pnName );

/**
 * \brief Display fritzfon preferences window
 * \param psParent parent widget
 */
static void fritzfonPreferences( GtkWidget *psParent ) {
	GtkWidget *psDialog = gtk_dialog_new_with_buttons( _( "FritzFon Preferences" ), GTK_WINDOW( psParent ), 0, GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE, NULL );
	GtkWidget *psCheckBox = gtk_check_button_new_with_label( _( "Swap first/lastname" ) );
	GtkWidget *psBox;

	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( psCheckBox ), fritzfonGetSwap() );
	psBox = gtk_dialog_get_content_area( GTK_DIALOG( psDialog ) );
	gtk_box_pack_start( GTK_BOX( psBox ), psCheckBox, FALSE, TRUE, 15 );
	gtk_widget_set_size_request( psDialog, 300, 80 );
	gtk_widget_show( GTK_WIDGET( psCheckBox ) );

	gtk_window_set_position( GTK_WINDOW( psDialog ), GTK_WIN_POS_CENTER_ON_PARENT );
	gtk_dialog_run( GTK_DIALOG( psDialog ) );

	prefsAddNone( getActiveProfile(), "/plugins/fritzfon" );

	fritzfonSetSwap( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( psCheckBox ) ) );
	gtk_widget_destroy( psDialog );
	SavePreferences( getActiveProfile() );

	freePersons();
	fritzfonReadBook();
}

/** book definition */
static struct sAddressBook sFritzFon = {
	fritzfonReadBook,
	fritzfonWriteBook,
	PERSON_FIRST_NAME | PERSON_LAST_NAME | PERSON_DISPLAY_NAME |
	PERSON_PRIVATE_PHONE | PERSON_PRIVATE_MOBILE |
	PERSON_BUSINESS_PHONE,
	PERSON_FIRST_NAME | PERSON_LAST_NAME | PERSON_DISPLAY_NAME |
	PERSON_PRIVATE_PHONE | PERSON_PRIVATE_MOBILE |
	PERSON_BUSINESS_PHONE,
	fritzfonPreferences
};

MODULE_INIT( PLUGIN_TYPE_BOOK, _( "FritzFon Addressbook" ), &sFritzFon, NULL, NULL );
