/*******************************************************************
 * Fritz Fun                                                       *
 * Created by Jan-Michael Brummer                                  *
 * All parts are distributed under the terms of GPLv2. See COPYING *
 *******************************************************************/

/**
 * \file ab_google.c
 * \brief google address book plugin
 */

#include <ffgtk.h>
#include <preferences.h>
#include <addressbook.h>

#include <gdata/gdata.h>

static GDataClientLoginAuthorizer *psAuthorizer = NULL;
static GDataContactsService *psService = NULL;

/** Get and set user name */
CREATE_STRING_PREFS( google, User, "/plugins/google/user" );

/**
 * \brief Get google password
 * \param psProfile profile structure
 * \return google password
 */
const gchar *googleGetPassword( struct sProfile *psProfile ) {
	struct sPassword *psPassword = NULL;

	psPassword = getDefaultPasswordPlugin( psProfile );
	if ( psPassword != NULL ) {
		return psPassword -> GetPassword( psProfile, "google" );
	}

	return prefsGetString( psProfile, "/plugins/google/password" );
}

/**
 * \brief Set google password
 * \param psProfile profile structure
 * \param pnPassword password
 */
void googleSetPassword( struct sProfile *psProfile, const gchar *pnPassword ) {
	struct sPassword *psPassword = NULL;

	psPassword = getDefaultPasswordPlugin( psProfile );
	if ( psPassword != NULL ) {
		psPassword -> SetPassword( psProfile, "google", pnPassword );
		return;
	}

	prefsSetString( psProfile, "/plugins/google/password", pnPassword );
}

/**
 * \brief Initialize gdata and authenticate user
 * \return error code, 0 on success otherwise error
 */
static int googleInit( void ) {
	struct sProfile *psProfile = getActiveProfile();
	gboolean bResult;
	GError *psError = NULL;

	psAuthorizer = gdata_client_login_authorizer_new( "ffgtk", GDATA_TYPE_CONTACTS_SERVICE );
	if ( psAuthorizer == NULL ) {
		Debug( KERN_DEBUG, "Could not create authorizer\n" );
		return -1;
	}

	psService = gdata_contacts_service_new( GDATA_AUTHORIZER( psAuthorizer ) );
	if ( psService == NULL ) {
		Debug( KERN_DEBUG, "Could not contacts service\n" );
		g_object_unref( psAuthorizer );
		return -2;
	}

	bResult = gdata_client_login_authorizer_authenticate( psAuthorizer, googleGetUser( psProfile ), googleGetPassword( psProfile ), NULL, &psError );
	if ( bResult == FALSE ) {
		Debug( KERN_DEBUG, "Could not authenticate: Wrong user data?\n" );
		if ( psError != NULL ) {
			Debug( KERN_DEBUG, "Error: %s\n", psError -> message );
		}
		g_object_unref( psService );
		g_object_unref( psAuthorizer );
		return -3;
	}

	return 0;
}

/**
 * \brief Deinitialize gdata
 * \return error code
 */
static int googleDeinit( void ) {
	if ( psService != NULL ) {
		g_object_unref( psService );
		psService = NULL;
	}
	if ( psAuthorizer != NULL ) {
		g_object_unref( psAuthorizer );
		psAuthorizer = NULL;
	}

	return 0;
}

/**
 * \brief Read address book
 * \return error code
 */
static int googleReadBook( void ) {
	GDataQuery *psQuery;
	GDataFeed *psFeed;
	GList *psList;
	GError *psError = NULL;
	int nChunkSize = 16;
	int nResult = 0;

	if ( googleInit() ) {
		Debug( KERN_WARNING, "googleInit() failed!\n" );
		return -1;
	}

	psQuery = GDATA_QUERY( gdata_contacts_query_new_with_limits( NULL, 1, nChunkSize ) );
	if ( psQuery == NULL ) {
		Debug( KERN_WARNING, "Contact query failed\n" );
		googleDeinit();
		return -2;
	}

	do {
		psFeed = gdata_contacts_service_query_contacts( psService, psQuery, NULL, NULL, NULL, &psError );
		if ( psFeed == NULL ) {
			Debug( KERN_WARNING, "Query contacts failed\n" );
			if ( psError != NULL ) {
				Debug( KERN_WARNING, "Error: %s\n", psError -> message );
				g_error_free( psError );
			}

			return -1;
		}

		psList = gdata_feed_get_entries( psFeed );
		nResult = psList ? g_list_length( psList ) : 0;

		for ( psList = gdata_feed_get_entries( psFeed ); psList != NULL; psList = psList -> next ) {
			GDataContactsContact *psContact;
			GDataGDName *psName = NULL;
			GList *psNumberList = NULL;
			GList *psAddressList = NULL;
			const gchar *pnId;
			GHashTable *psTable;
			guint8 *pnPhoto;
			gsize nPhotoLen;
			gchar *pnPhotoType;
			GDataGDOrganization *psOrganization;

			if ( psList -> data == NULL ) {
				Debug( KERN_WARNING, "Strange... data == NULL\n" );
				continue;
			}

			psContact = GDATA_CONTACTS_CONTACT( psList -> data );
			if ( psContact == NULL ) {
				Debug( KERN_WARNING, "psContact == NULL\n" );
				return -2;
			}

			pnId = gdata_entry_get_id( GDATA_ENTRY( psList -> data ) );
			if ( pnId == NULL ) {
				Debug( KERN_WARNING, "pnId == NULL\n" );
				return -2;
			}

			psTable = g_hash_table_new( NULL, NULL );
			if ( psTable == NULL ) {
				Debug( KERN_WARNING, "psTable == NULL\n" );
				return -2;
			}

			AddInfo( psTable, PERSON_ID, pnId );

			psName = gdata_contacts_contact_get_name( psContact );
			if ( psName != NULL ) {
				const gchar *pnFirstName = gdata_gd_name_get_given_name( psName );
				const gchar *pnLastName = gdata_gd_name_get_family_name( psName );

				if ( pnFirstName != NULL ) {
					AddInfo( psTable, PERSON_FIRST_NAME, pnFirstName );
				}

				if ( pnLastName != NULL ) {
					AddInfo( psTable, PERSON_LAST_NAME, pnLastName );
				}

				AddInfo( psTable, PERSON_DISPLAY_NAME, gdata_gd_name_get_full_name( psName ) );
				AddInfo( psTable, PERSON_TITLE, gdata_gd_name_get_prefix( psName ) );
			}

			psNumberList = gdata_contacts_contact_get_phone_numbers( psContact );
			while ( psNumberList != NULL && psNumberList -> data != NULL ) {
				GDataGDPhoneNumber *psNumber = psNumberList -> data;
				const gchar *pnType = gdata_gd_phone_number_get_relation_type( psNumber );
				const gchar *pnNumber = gdata_gd_phone_number_get_number( psNumber );

				if ( pnType == NULL ) {
					Debug( KERN_WARNING, "pnType == NULL\n" );
					break;
				}
				if ( pnNumber == NULL ) {
					Debug( KERN_WARNING, "pnNumber == NULL\n" );
					break;
				}

				if ( strcmp( pnType, GDATA_GD_PHONE_NUMBER_WORK ) == 0 ) {
					AddInfo( psTable, PERSON_BUSINESS_PHONE, pnNumber );
				} else if ( strcmp( pnType, GDATA_GD_PHONE_NUMBER_HOME ) == 0 ) {
					AddInfo( psTable, PERSON_PRIVATE_PHONE, pnNumber );
				} else if ( strcmp( pnType, GDATA_GD_PHONE_NUMBER_MOBILE ) == 0 ) {
					AddInfo( psTable, PERSON_PRIVATE_MOBILE, pnNumber );
				} else if ( strcmp( pnType, GDATA_GD_PHONE_NUMBER_HOME_FAX ) == 0 ) {
					AddInfo( psTable, PERSON_PRIVATE_FAX, pnNumber );
				} else if ( strcmp( pnType, GDATA_GD_PHONE_NUMBER_WORK_FAX ) == 0 ) {
					AddInfo( psTable, PERSON_BUSINESS_FAX, pnNumber );
				}

				psNumberList = psNumberList -> next;
			}

			psAddressList = gdata_contacts_contact_get_postal_addresses( psContact );
			while ( psAddressList != NULL && psAddressList -> data != NULL ) {
				GDataGDPostalAddress *psAddress = psAddressList -> data;
				const gchar *pnType = gdata_gd_postal_address_get_relation_type( psAddress );

				if ( pnType == NULL ) {
					Debug( KERN_WARNING, "pnType == NULL\n" );
					break;
				}

				if ( !strcmp( pnType, GDATA_GD_POSTAL_ADDRESS_WORK ) ) {
					AddInfo( psTable, PERSON_BUSINESS_STREET, gdata_gd_postal_address_get_street( psAddress ) );
					AddInfo( psTable, PERSON_BUSINESS_CITY, gdata_gd_postal_address_get_city( psAddress ) );
					AddInfo( psTable, PERSON_BUSINESS_COUNTRY, gdata_gd_postal_address_get_country( psAddress ) );
					AddInfo( psTable, PERSON_BUSINESS_ZIPCODE, gdata_gd_postal_address_get_postcode( psAddress ) );
				} else if ( !strcmp( pnType, GDATA_GD_POSTAL_ADDRESS_HOME ) ) {
					AddInfo( psTable, PERSON_PRIVATE_STREET, gdata_gd_postal_address_get_street( psAddress ) );
					AddInfo( psTable, PERSON_PRIVATE_CITY, gdata_gd_postal_address_get_city( psAddress ) );
					AddInfo( psTable, PERSON_PRIVATE_COUNTRY, gdata_gd_postal_address_get_country( psAddress ) );
					AddInfo( psTable, PERSON_PRIVATE_ZIPCODE, gdata_gd_postal_address_get_postcode( psAddress ) );
				}
				psAddressList = psAddressList -> next;
			}

			pnPhoto = gdata_contacts_contact_get_photo( psContact, psService, &nPhotoLen, &pnPhotoType, NULL, NULL );
			if ( pnPhoto != NULL ) {
				GdkPixbufLoader *psLoader;

				psLoader = gdk_pixbuf_loader_new();
				if ( gdk_pixbuf_loader_write( psLoader, pnPhoto, nPhotoLen, NULL ) ) {
					AddInfo( psTable, PERSON_IMAGE, ( const gchar * ) gdk_pixbuf_loader_get_pixbuf( psLoader ) );
				}
				gdk_pixbuf_loader_close( psLoader, NULL );
			}

			psOrganization = gdata_contacts_contact_get_primary_organization( psContact );
			if ( psOrganization != NULL ) {
				AddInfo( psTable, PERSON_COMPANY, gdata_gd_organization_get_name( psOrganization ) );
			}

			AddPerson( psTable, FALSE );

			g_hash_table_destroy( psTable );
		}

		gdata_query_next_page( GDATA_QUERY( psQuery ) );
	} while ( nResult == nChunkSize );

	Debug( KERN_DEBUG, "done\n" );
	return 0;
}

void googleSetImage( GDataContactsContact *psContact, struct sPerson *psPerson ) {
	if ( psPerson -> pnNewImage != NULL ) {
		gchar *pnData;
		gsize nLength;

		if ( g_file_get_contents( psPerson -> pnNewImage, &pnData, &nLength, NULL ) ) {
			gdata_contacts_contact_set_photo( psContact, psService, ( guint8 * ) pnData, nLength, "image/jpeg", NULL, NULL );
		}
	} else if ( psPerson -> psImage == NULL ) {
		gdata_contacts_contact_set_photo( psContact, psService, NULL, 0, NULL, NULL, NULL );
	}
}

/**
 * \brief Save address book
 * \return error code
 */
int googleSaveBook( void ) {
	GList *psList;
	GError *psError = NULL;
	struct sProfile *psProfile = getActiveProfile();

	Debug( KERN_DEBUG, "started\n" );
	if ( googleInit() ) {
		Debug( KERN_DEBUG, "googleInit() failed!\n" );
		return -1;
	}

	for ( psList = psPersonsList; psList != NULL && psList -> data != NULL; psList = psList -> next ) {
		struct sPerson *psPerson = psList -> data;

		if ( psPerson -> nFlags == PERSON_FLAG_UNCHANGED ) {
			continue;
		}

		if ( psPerson -> nFlags & PERSON_FLAG_NEW ) {
			GDataGDName *psName;
			GDataGDPhoneNumber *psPhone;
			GDataContactsContact *psContact;
			GDataGDPostalAddress *psAddress;
			GDataGDOrganization *psOrganization;
			gchar *pnIdUrl = NULL;

			Debug( KERN_DEBUG, "Adding new person\n" );

			/* Create new contact */
			psContact = gdata_contacts_contact_new( NULL );
			g_assert( GDATA_IS_CONTACTS_CONTACT( psContact ) );

			/* Add group id */
			pnIdUrl = g_strdup_printf( "http://www.google.com/m8/feeds/groups/%s/base/6", googleGetUser( psProfile ) );
			gdata_contacts_contact_add_group( psContact, pnIdUrl );
			g_free( pnIdUrl );

			/* Set display name */
			gdata_entry_set_title( GDATA_ENTRY( psContact ), psPerson -> pnDisplayName );

			/* Set name */
			psName = gdata_gd_name_new( psPerson -> pnFirstName, psPerson -> pnLastName );
			if ( psPerson -> pnTitle != NULL && strlen( psPerson -> pnTitle ) > 0 ) {
				gdata_gd_name_set_prefix( psName, psPerson -> pnTitle );
			}
			gdata_contacts_contact_set_name( psContact, psName );
			g_object_unref( psName );

			/* Set company */
			if ( psPerson -> pnCompany != NULL ) {
				psOrganization = gdata_gd_organization_new( psPerson -> pnCompany, NULL, NULL, NULL, TRUE );

				gdata_contacts_contact_add_organization( psContact, psOrganization );

				g_object_unref( psOrganization );
			}

			/* Add private phone */
			if ( psPerson -> pnPrivatePhone != NULL && strlen( psPerson -> pnPrivatePhone ) > 0 ) {
				psPhone = gdata_gd_phone_number_new( psPerson -> pnPrivatePhone, GDATA_GD_PHONE_NUMBER_HOME, NULL, NULL, TRUE );
				gdata_contacts_contact_add_phone_number( psContact, psPhone );
				g_object_unref( psPhone );
			}

			/* Add business phone */
			if ( psPerson -> pnBusinessPhone != NULL && strlen( psPerson -> pnBusinessPhone ) > 0 ) {
				psPhone = gdata_gd_phone_number_new( psPerson -> pnBusinessPhone, GDATA_GD_PHONE_NUMBER_WORK, NULL, NULL, TRUE );
				gdata_contacts_contact_add_phone_number( psContact, psPhone );
				g_object_unref( psPhone );
			}

			/* Add private mobile */
			if ( psPerson -> pnPrivateMobile != NULL && strlen( psPerson -> pnPrivateMobile ) > 0 ) {
				psPhone = gdata_gd_phone_number_new( psPerson -> pnPrivateMobile, GDATA_GD_PHONE_NUMBER_MOBILE, NULL, NULL, TRUE );
				gdata_contacts_contact_add_phone_number( psContact, psPhone );
				g_object_unref( psPhone );
			}

			/* Add private fax */
			if ( psPerson -> pnPrivateFax != NULL && strlen( psPerson -> pnPrivateFax ) > 0 ) {
				psPhone = gdata_gd_phone_number_new( psPerson -> pnPrivateFax, GDATA_GD_PHONE_NUMBER_HOME_FAX, NULL, NULL, TRUE );
				gdata_contacts_contact_add_phone_number( psContact, psPhone );
				g_object_unref( psPhone );
			}

			/* Add business fax */
			if ( psPerson -> pnBusinessFax != NULL && strlen( psPerson -> pnBusinessFax ) > 0 ) {
				psPhone = gdata_gd_phone_number_new( psPerson -> pnBusinessFax, GDATA_GD_PHONE_NUMBER_WORK_FAX, NULL, NULL, TRUE );
				gdata_contacts_contact_add_phone_number( psContact, psPhone );
				g_object_unref( psPhone );
			}

			/* Add business address */
			psAddress = gdata_gd_postal_address_new( GDATA_GD_POSTAL_ADDRESS_WORK, NULL, TRUE );
			if ( psAddress != NULL ) {
				if ( psPerson -> pnBusinessStreet != NULL && strlen( psPerson -> pnBusinessStreet ) > 0 ) {
					gdata_gd_postal_address_set_street( psAddress, psPerson -> pnBusinessStreet );
				}
				if ( psPerson -> pnBusinessCity != NULL && strlen( psPerson -> pnBusinessCity ) > 0 ) {
					gdata_gd_postal_address_set_city( psAddress, psPerson -> pnBusinessCity );
				}
				if ( psPerson -> pnBusinessCountry != NULL && strlen( psPerson -> pnBusinessCountry ) > 0 ) {
					gdata_gd_postal_address_set_country( psAddress, psPerson -> pnBusinessCountry, NULL );
				}
				if ( psPerson -> pnBusinessZipCode != NULL && strlen( psPerson -> pnBusinessZipCode ) > 0 ) {
					gdata_gd_postal_address_set_postcode( psAddress, psPerson -> pnBusinessZipCode );
				}

				gdata_contacts_contact_add_postal_address( psContact, psAddress );

				g_object_unref( psAddress );
			}

			/* Add private address */
			psAddress = gdata_gd_postal_address_new( GDATA_GD_POSTAL_ADDRESS_HOME, NULL, TRUE );
			if ( psAddress != NULL ) {
				if ( psPerson -> pnPrivateStreet != NULL && strlen( psPerson -> pnPrivateStreet ) > 0 ) {
					gdata_gd_postal_address_set_street( psAddress, psPerson -> pnPrivateStreet );
				}
				if ( psPerson -> pnPrivateCity != NULL && strlen( psPerson -> pnPrivateCity ) > 0 ) {
					gdata_gd_postal_address_set_city( psAddress, psPerson -> pnPrivateCity );
				}
				if ( psPerson -> pnPrivateCountry != NULL && strlen( psPerson -> pnPrivateCountry ) > 0 ) {
					gdata_gd_postal_address_set_country( psAddress, psPerson -> pnPrivateCountry, NULL );
				}
				if ( psPerson -> pnPrivateZipCode != NULL && strlen( psPerson -> pnPrivateZipCode ) > 0 ) {
					gdata_gd_postal_address_set_postcode( psAddress, psPerson -> pnPrivateZipCode );
				}

				gdata_contacts_contact_add_postal_address( psContact, psAddress );

				g_object_unref( psAddress );
			}

			/* Insert contact */
			psContact = gdata_contacts_service_insert_contact( GDATA_CONTACTS_SERVICE( psService ), psContact, NULL, &psError );

			/* Now add image */
			googleSetImage( psContact, psPerson );

			/* Update person id */
			psPerson -> pnId = ( gchar * ) gdata_entry_get_id( GDATA_ENTRY( psContact ) );
		}

		if ( psPerson -> nFlags & PERSON_FLAG_DELETED ) {
			GDataEntry *psEntry;

			/* Query contact */
			psEntry = gdata_service_query_single_entry( GDATA_SERVICE( psService ), gdata_contacts_service_get_primary_authorization_domain(), psPerson -> pnId, NULL, GDATA_TYPE_CONTACTS_CONTACT, NULL, NULL );
			if ( psEntry != NULL ) {
				/* Delete valid entry */
				gdata_service_delete_entry( GDATA_SERVICE( psService ), gdata_contacts_service_get_primary_authorization_domain(), psEntry, NULL, NULL );

				g_object_unref( psEntry );
			}
		}

		if ( psPerson -> nFlags & PERSON_FLAG_CHANGED ) {
			GDataEntry *psEntry;
			GDataContactsContact *psNewContact;
			GDataContactsContact *psContact;
			GDataGDName *psName;
			GDataGDPostalAddress *psAddress;
			GDataGDOrganization *psOrganization;
			GList *psNumberList;
			GError *psError = NULL;
			gboolean bBusinessPhone = FALSE;
			gboolean bBusinessFax = FALSE;
			gboolean bPrivateFax = FALSE;
			gboolean bPrivatePhone = FALSE;
			gboolean bPrivateMobile = FALSE;

			Debug( KERN_DEBUG, "Changed entry\n" );

			/* Query contact */
			psEntry = gdata_service_query_single_entry( GDATA_SERVICE( psService ), gdata_contacts_service_get_primary_authorization_domain(), psPerson -> pnId, NULL, GDATA_TYPE_CONTACTS_CONTACT, NULL, NULL );
			if ( psEntry == NULL ) {
				Debug( KERN_DEBUG, "psEntry is NULL\n");
				continue;
			}
			g_assert( GDATA_IS_CONTACTS_CONTACT( psEntry ) );

			/* Convert to contact */
			psContact = GDATA_CONTACTS_CONTACT( psEntry );

			/* Update title */
			gchar *pnDisplayName = g_strdup_printf( "%s %s", psPerson -> pnFirstName, psPerson -> pnLastName );
			gdata_entry_set_title( GDATA_ENTRY( psContact ), pnDisplayName );

			/* Get name information and update name information */
			psName = gdata_contacts_contact_get_name( psContact );
			if ( psName != NULL ) {
				if ( psPerson -> pnFirstName != NULL && strlen( psPerson -> pnFirstName ) > 0 ) {
					gdata_gd_name_set_given_name( psName, psPerson -> pnFirstName );
				} else {
				}

				if ( psPerson -> pnLastName != NULL && strlen( psPerson -> pnLastName ) > 0 ) {
					gdata_gd_name_set_family_name( psName, psPerson -> pnLastName );
				}
				if ( pnDisplayName != NULL && strlen( psPerson -> pnDisplayName ) > 0 ) {
					gdata_gd_name_set_full_name( psName, pnDisplayName );
				}
				if ( psPerson -> pnTitle != NULL && strlen( psPerson -> pnTitle ) > 0 ) {
					gdata_gd_name_set_prefix( psName, psPerson -> pnTitle );
				}
				Debug( KERN_DEBUG, "Set names done\n" );
				//g_object_unref( psName );
			}

			/* Update company */
			if ( psPerson -> pnCompany != NULL && strlen( psPerson -> pnCompany ) > 0 ) {
				psOrganization = gdata_contacts_contact_get_primary_organization( psContact );

				if ( psOrganization == NULL ) {
					psOrganization = gdata_gd_organization_new( psPerson -> pnCompany, NULL, NULL, NULL, TRUE );

					gdata_contacts_contact_add_organization( psContact, psOrganization );
				} else {
					gdata_gd_organization_set_name( psOrganization, psPerson -> pnCompany );
				}

				//g_object_unref( psOrganization );
			}

			/* Set numbers */
			psNumberList = gdata_contacts_contact_get_phone_numbers( psContact );
			while ( psNumberList != NULL && psNumberList -> data != NULL ) {
				GDataGDPhoneNumber *psNumber = psNumberList -> data;

				if ( !strcmp( gdata_gd_phone_number_get_relation_type( psNumber ), GDATA_GD_PHONE_NUMBER_WORK ) ) {
					bBusinessPhone = TRUE;
					if ( strlen( psPerson -> pnBusinessPhone ) <= 0 ) {
						psNumberList = g_list_remove( psNumberList, psNumber );
						continue;
					} else {
						gdata_gd_phone_number_set_number( psNumber, psPerson -> pnBusinessPhone );
					}
				} else if ( !strcmp( gdata_gd_phone_number_get_relation_type( psNumber ), GDATA_GD_PHONE_NUMBER_HOME ) ) {
					bPrivatePhone = TRUE;
					if ( strlen( psPerson -> pnPrivatePhone ) <= 0 ) {
						psNumberList = g_list_remove( psNumberList, psNumber );
						continue;
					} else {
						gdata_gd_phone_number_set_number( psNumber, psPerson -> pnPrivatePhone );
					}
				} else if ( !strcmp( gdata_gd_phone_number_get_relation_type( psNumber ), GDATA_GD_PHONE_NUMBER_MOBILE ) ) {
					bPrivateMobile = TRUE;
					if ( strlen( psPerson -> pnPrivateMobile ) <= 0 ) {
						psNumberList = g_list_remove( psNumberList, psNumber );
						continue;
					} else {
						gdata_gd_phone_number_set_number( psNumber, psPerson -> pnPrivateMobile );
					}
				} else if ( !strcmp( gdata_gd_phone_number_get_relation_type( psNumber ), GDATA_GD_PHONE_NUMBER_HOME_FAX ) ) {
					bPrivateFax = TRUE;
					if ( strlen( psPerson -> pnPrivateFax ) <= 0 ) {
						psNumberList = g_list_remove( psNumberList, psNumber );
						continue;
					} else {
						gdata_gd_phone_number_set_number( psNumber, psPerson -> pnPrivateFax );
					}
				} else if ( !strcmp( gdata_gd_phone_number_get_relation_type( psNumber ), GDATA_GD_PHONE_NUMBER_WORK_FAX ) ) {
					bBusinessFax = TRUE;
					if ( strlen( psPerson -> pnBusinessFax ) <= 0 ) {
						psNumberList = g_list_remove( psNumberList, psNumber );
						continue;
					} else {
						gdata_gd_phone_number_set_number( psNumber, psPerson -> pnBusinessFax );
					}
				}
				//g_object_unref( psNumber );

				if ( psNumberList != NULL ) {
					psNumberList = psNumberList -> next;
				}
			}
			if ( bBusinessPhone == FALSE && strlen( psPerson -> pnBusinessPhone ) > 0 ) {
				GDataGDPhoneNumber *psNumber = gdata_gd_phone_number_new( psPerson -> pnBusinessPhone, GDATA_GD_PHONE_NUMBER_WORK, NULL, NULL, FALSE );

				gdata_contacts_contact_add_phone_number( psContact, psNumber );
			}

			if ( bBusinessFax == FALSE && strlen( psPerson -> pnBusinessFax ) > 0 ) {
				GDataGDPhoneNumber *psNumber = gdata_gd_phone_number_new( psPerson -> pnBusinessFax, GDATA_GD_PHONE_NUMBER_WORK_FAX, NULL, NULL, FALSE );

				gdata_contacts_contact_add_phone_number( psContact, psNumber );
			}

			if ( bPrivatePhone == FALSE && strlen( psPerson -> pnPrivatePhone ) > 0 ) {
				GDataGDPhoneNumber *psNumber = gdata_gd_phone_number_new( psPerson -> pnPrivatePhone, GDATA_GD_PHONE_NUMBER_HOME, NULL, NULL, FALSE );

				gdata_contacts_contact_add_phone_number( psContact, psNumber );
			}

			if ( bPrivateFax == FALSE && strlen( psPerson -> pnPrivateFax ) > 0 ) {
				GDataGDPhoneNumber *psNumber = gdata_gd_phone_number_new( psPerson -> pnPrivateFax, GDATA_GD_PHONE_NUMBER_HOME_FAX, NULL, NULL, FALSE );

				gdata_contacts_contact_add_phone_number( psContact, psNumber );
			}

			if ( bPrivateMobile == FALSE && strlen( psPerson -> pnPrivateMobile ) > 0 ) {
				GDataGDPhoneNumber *psNumber = gdata_gd_phone_number_new( psPerson -> pnPrivateMobile, GDATA_GD_PHONE_NUMBER_MOBILE, NULL, NULL, FALSE );

				gdata_contacts_contact_add_phone_number( psContact, psNumber );
			}

			Debug( KERN_DEBUG, "Set numbers done\n" );

			psAddress = gdata_gd_postal_address_new( GDATA_GD_POSTAL_ADDRESS_WORK, NULL, TRUE );
			if ( psPerson -> pnBusinessStreet != NULL && strlen( psPerson -> pnBusinessStreet ) > 0 ) {
				gdata_gd_postal_address_set_street( psAddress, psPerson -> pnBusinessStreet );
			}
			if ( psPerson -> pnBusinessCity != NULL && strlen( psPerson -> pnBusinessCity ) > 0 ) {
				gdata_gd_postal_address_set_city( psAddress, psPerson -> pnBusinessCity );
			}
			if ( psPerson -> pnBusinessCountry != NULL && strlen( psPerson -> pnBusinessCountry ) > 0 ) {
				gdata_gd_postal_address_set_country( psAddress, psPerson -> pnBusinessCountry, NULL );
			}
			if ( psPerson -> pnBusinessZipCode != NULL && strlen( psPerson -> pnBusinessZipCode ) > 0 ) {
				gdata_gd_postal_address_set_postcode( psAddress, psPerson -> pnBusinessZipCode );
			}

			psAddress = gdata_gd_postal_address_new( GDATA_GD_POSTAL_ADDRESS_HOME, NULL, TRUE );
			if ( psPerson -> pnPrivateStreet != NULL && strlen( psPerson -> pnPrivateStreet ) > 0 ) {
				gdata_gd_postal_address_set_street( psAddress, psPerson -> pnPrivateStreet );
			}
			if ( psPerson -> pnPrivateCity != NULL  && strlen( psPerson -> pnPrivateCity ) > 0) {
				gdata_gd_postal_address_set_city( psAddress, psPerson -> pnPrivateCity );
			}
			if ( psPerson -> pnPrivateCountry != NULL  && strlen( psPerson -> pnPrivateCountry ) > 0) {
				gdata_gd_postal_address_set_country( psAddress, psPerson -> pnPrivateCountry, NULL );
			}
			if ( psPerson -> pnPrivateZipCode != NULL && strlen( psPerson -> pnPrivateZipCode ) > 0 ) {
				gdata_gd_postal_address_set_postcode( psAddress, psPerson -> pnPrivateZipCode );
			}

			googleSetImage( psContact, psPerson );

			Debug( KERN_DEBUG, "Update entry\n" );
			psNewContact = GDATA_CONTACTS_CONTACT( gdata_service_update_entry( GDATA_SERVICE( psService ), gdata_contacts_service_get_primary_authorization_domain(), GDATA_ENTRY( psContact ), NULL, &psError ) );
			if ( psError != NULL ) {
				Debug( KERN_WARNING, "Could not update contact entry: %s\n", psError -> message );
				g_error_free( psError );
				psError = NULL;
			} else {
				Debug( KERN_DEBUG, "Update entry done\n" );
				if ( psNewContact ) {
					Debug( KERN_DEBUG, "Unref\n" );
					g_object_unref( psNewContact );
				}
			}
		}

		psPerson -> nFlags = PERSON_FLAG_UNCHANGED;
	}
	return 0;
}

/**
 * \brief Display google preferences window
 * \param psParent parent widget
 */
static void googlePreferences( GtkWidget *psParent ) {
	GtkWidget *psDialog;
	GtkWidget *psBox;
	GtkWidget *psLabel;
	GtkWidget *psUser;
	GtkWidget *psPassword;
	struct sProfile *psProfile = getActiveProfile();

	psDialog = gtk_dialog_new_with_buttons( _( "Google Preferences" ), GTK_WINDOW( psParent ), 0, GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE, NULL );
	psBox = gtk_dialog_get_content_area( GTK_DIALOG( psDialog ) );

	psLabel = gtk_label_new( "Name:" );
	gtk_box_pack_start( GTK_BOX( psBox ), psLabel, FALSE, TRUE, 10 );
	psUser = gtk_entry_new();
	gtk_entry_set_text( GTK_ENTRY( psUser ), googleGetUser( psProfile ) );
	gtk_box_pack_start( GTK_BOX( psBox ), psUser, FALSE, TRUE, 10 );
	psLabel = gtk_label_new( "Password:" );
	gtk_box_pack_start( GTK_BOX( psBox ), psLabel, FALSE, TRUE, 10 );
	psPassword = gtk_entry_new();
	gtk_entry_set_visibility( GTK_ENTRY( psPassword ), FALSE );
	gtk_entry_set_text( GTK_ENTRY( psPassword ), googleGetPassword( psProfile ) );
	gtk_box_pack_start( GTK_BOX( psBox ), psPassword, FALSE, TRUE, 10 );

	gtk_widget_show_all( GTK_WIDGET( psDialog ) );
	gtk_widget_set_size_request( psDialog, 300, 220 );
	gtk_window_set_position( GTK_WINDOW( psDialog ), GTK_WIN_POS_CENTER );
	gtk_dialog_run( GTK_DIALOG( psDialog ) );

	prefsAddNone( psProfile, "/plugins/google" );

	googleSetUser( psProfile, gtk_entry_get_text( GTK_ENTRY( psUser ) ) );
	googleSetPassword( psProfile, gtk_entry_get_text( GTK_ENTRY( psPassword ) ) );

	gtk_widget_destroy( psDialog );
	SavePreferences( psProfile );
}

/** book definition */
struct sAddressBook sGoogle = {
	googleReadBook,
	googleSaveBook,
	PERSON_FIRST_NAME | PERSON_LAST_NAME | PERSON_DISPLAY_NAME | PERSON_COMPANY |
	PERSON_TITLE | PERSON_IMAGE | PERSON_PRIVATE_PHONE | PERSON_PRIVATE_MOBILE |
	PERSON_PRIVATE_FAX | PERSON_PRIVATE_STREET | PERSON_PRIVATE_CITY |
	PERSON_PRIVATE_ZIPCODE | PERSON_PRIVATE_COUNTRY | PERSON_BUSINESS_PHONE |
	PERSON_BUSINESS_FAX | PERSON_BUSINESS_STREET | PERSON_BUSINESS_CITY |
	PERSON_BUSINESS_ZIPCODE | PERSON_BUSINESS_COUNTRY,
	PERSON_FIRST_NAME | PERSON_LAST_NAME | PERSON_DISPLAY_NAME | PERSON_COMPANY |
	PERSON_TITLE | PERSON_IMAGE | PERSON_PRIVATE_PHONE | PERSON_PRIVATE_MOBILE |
	PERSON_PRIVATE_FAX | PERSON_PRIVATE_STREET | PERSON_PRIVATE_CITY |
	PERSON_PRIVATE_ZIPCODE | PERSON_PRIVATE_COUNTRY | PERSON_BUSINESS_PHONE |
	PERSON_BUSINESS_FAX | PERSON_BUSINESS_STREET | PERSON_BUSINESS_CITY |
	PERSON_BUSINESS_ZIPCODE | PERSON_BUSINESS_COUNTRY,
	googlePreferences
};

MODULE_INIT( PLUGIN_TYPE_BOOK, _( "Google Contacts (BETA)" ), &sGoogle, NULL, NULL );
